import unittest
import warnings

import nextnanopy.outputs as outputs
from nextnanopy.utils.datasets import default_unit
from os.path import join
import os

TESTS_DIR = os.path.dirname(os.path.abspath(__file__))
# folder_nnp = os.path.join(TESTS_DIR, 'datafiles', 'nextnano++')
# folder_nn3 = os.path.join(TESTS_DIR, 'datafiles', 'nextnano3')
# folder_negf = os.path.join(TESTS_DIR, 'datafiles', 'nextnano.NEGF')
# folder_msb = os.path.join(TESTS_DIR, 'datafiles', 'nextnano.MSB')

folder_nnp = os.path.join("tests", 'datafiles', 'nextnano++')
folder_nn3 = os.path.join("tests", 'datafiles', 'nextnano3')
folder_negf = os.path.join("tests", 'datafiles', 'nextnano.NEGF')
folder_msb = os.path.join("tests", 'datafiles', 'nextnano.MSB')


class TestOutputs_nnp(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(join(folder_nnp, 'bandedges_1d.dat'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords['x'].name, 'x')
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 548)
        self.assertEqual(len(df.variables.keys()), 4)
        self.assertEqual(df.variables['Gamma'].name, 'Gamma')
        self.assertEqual(df.variables['Gamma'].unit, 'eV')
        self.assertEqual(df.variables['Gamma'].value.size, 548)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

        df = outputs.DataFile(join(folder_nnp, 'wf_occupation_1d.dat'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.coords['no.'].name, 'no.')
        self.assertEqual(df.coords['no.'].unit, default_unit)
        self.assertEqual(df.coords['no.'].value.size, 10)
        self.assertEqual(df.variables['Occupation'].name, 'Occupation')
        self.assertEqual(df.variables['Occupation'].unit, 'electrons/cm^2')
        self.assertEqual(df.variables['Occupation'].value.size, 10)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

        def test_dat_FirstVarFlag(self):
            df = outputs.DataFile(join(folder_nnp, 'bandedges_1d.dat'), product='nextnano++', FirstVarIsCoordFlag = False)
            self.assertEqual(len(df.coords.keys()), 0)
            self.assertEqual(len(df.variables.keys()), 5)
            self.assertEqual(df.variables['x'].name, 'x')
            self.assertEqual(df.variables['x'].unit, 'nm')
            self.assertEqual(df.variables['x'].value.size, 548)
            self.assertEqual(df.variables['Gamma'].name, 'Gamma')
            self.assertEqual(df.variables['Gamma'].unit, 'eV')
            self.assertEqual(df.variables['Gamma'].value.size, 548)
            self.assertEqual(df.metadata['ndim'], 1)
            self.assertEqual(df.metadata['dkeys'], [0])

            df = outputs.DataFile(join(folder_nnp, 'wf_occupation_1d.dat'), product='nextnano++', FirstVarIsCoordFlag = True)
            self.assertEqual(len(df.coords.keys()), 1)
            self.assertEqual(len(df.variables.keys()), 1)
            self.assertEqual(df.coords['no.'].name, 'no.')
            self.assertEqual(df.coords['no.'].unit, default_unit)
            self.assertEqual(df.coords['no.'].value.size, 10)
            self.assertEqual(df.variables['Occupation'].name, 'Occupation')
            self.assertEqual(df.variables['Occupation'].unit, 'electrons/cm^2')
            self.assertEqual(df.variables['Occupation'].value.size, 10)
            self.assertEqual(df.metadata['ndim'], 1)
            self.assertEqual(df.metadata['dkeys'], [0])

        #test creating the document
        self.assertRaises(NotImplementedError, df.export, filename='TESTEXPORT.dat', format='dat')

    def test_avs(self):
        df = outputs.DataFile(join(folder_nnp, 'bandedges_2d.fld'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 164)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, 'nm')
        self.assertEqual(df.coords['y'].value.size, 79)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(len(df.variables.keys()), 2)
        self.assertEqual(df.variables['Gamma'].name, 'Gamma')
        self.assertEqual(df.variables['Gamma'].unit, 'eV')
        self.assertEqual(df.variables['Gamma'].value.shape, (164, 79))
        self.assertEqual(df.metadata['ndim'], 2)
        self.assertEqual(df.metadata['dims'], [164, 79])
        self.assertEqual(df.metadata['field'], 'rectilinear')

        # test creating the document
        self.assertRaises(NotImplementedError, df.export, filename = 'TESTEXPORT2D.fld', format='AVSAscii')

    def test_avs3D(self):
        df = outputs.DataFile(join(folder_nnp, 'potential.fld'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ['x', 'y', 'z'])
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 11)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, 'nm')
        self.assertEqual(df.coords['y'].value.size, 11)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(df.coords['z'].unit, 'nm')
        self.assertEqual(df.coords['z'].value.size, 76)
        self.assertEqual(df.coords['z'].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['potential'].name, 'potential')
        self.assertEqual(df.variables['potential'].unit, default_unit)
        self.assertEqual(df.variables['potential'].value.shape, (11, 11, 76))
        self.assertEqual(df.metadata['ndim'], 3)
        self.assertEqual(df.metadata['dims'], [11, 11, 76])
        self.assertEqual(df.metadata['field'], 'rectilinear')

        # test creating the document
        self.assertRaises(NotImplementedError, df.export, filename = 'TESTEXPORT3D.fld', format='AVSAscii')

    def test_vtr_potential(self):
        df = outputs.DataFile(join(folder_nnp, 'potential.vtr'), product='nextnano++')

        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ['x', 'y', 'z'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 11)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 11)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(df.coords['z'].unit, default_unit)
        self.assertEqual(df.coords['z'].value.size, 76)
        self.assertEqual(df.coords['z'].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['potential'].name, 'potential')
        self.assertEqual(df.variables['potential'].unit, default_unit)
        self.assertEqual(df.variables['potential'].value.shape, (11, 11, 76))
        self.assertEqual(df.metadata, {})

    def test_vtr_bandedges(self):
        df = outputs.DataFile(join(folder_nnp, 'bandedges.vtr'), product='nextnano++')

        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ['x', 'y', 'z'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 11)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 11)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(df.coords['z'].unit, default_unit)
        self.assertEqual(df.coords['z'].value.size, 76)
        self.assertEqual(df.coords['z'].dim, 2)
        self.assertEqual(len(df.variables.keys()), 4)
        self.assertEqual(df.variables['Gamma'].name, 'Gamma')
        self.assertEqual(df.variables['Gamma'].unit, 'eV')
        self.assertEqual(df.variables['Gamma'].value.shape, (11, 11, 76))
        self.assertEqual(df.metadata, {})

    def test_txt(self):
        df = outputs.DataFile(join(folder_nnp, 'variables_input.txt'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 3)
        self.assertEqual(df.variables['var_1'].name, 'var_1')
        self.assertEqual(df.variables['var_1'].unit, default_unit)
        self.assertEqual(df.variables['var_1'].value, -47.5)
        self.assertEqual(df.variables['var_2'].value, 75)
        self.assertEqual(df.variables['string'].value, '"hello"')

        df = outputs.DataFile(join(folder_nnp, 'variables_database.txt'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 0)

        df = outputs.DataFile(join(folder_nnp, 'total_charges.txt'), product='nextnano++')
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 6)
        print(df.variables['fixed'].value)
        self.assertAlmostEqual(df.variables['fixed'].value, 0.0)
        self.assertEqual(df.variables[0].name, 'electron')
        self.assertEqual(df.variables[0].unit, 'e/cm^2')

        files = ['simulation_database.txt', 'simulation_info.txt', 'simulation_input.txt']
        for file in files:
            self.assertRaises(NotImplementedError, outputs.DataFile, join(folder_nnp, file), 'nextnano++')

    def test_rest(self):
        files = ['example.log', 'example.in']
        for file in files:
            self.assertRaises(NotImplementedError, outputs.DataFile, join(folder_nnp, file), 'nextnano++')

    def test_all_2D(self):
        files = [r'AvsAscii\bandedges.avs.fld',r'AvsAscii_one_file\bandedges.fld',r'AvsBinary\bandedges.avs.fld',r'AvsBinary_one_file\bandedges.fld',r'VTKAscii\bandedges.vtr']
        for file in files:
            filepath = os.path.join(folder_nnp, file)
            datafile = outputs.DataFile(filepath, product = 'nextnano++')
            self.assertEqual(len(datafile.coords),2)
            self.assertEqual(len(datafile.variables),6)

    #TODO add tests for loading Origin like file
    #TODO add tests for exporting binary data

class TestOutputs_nn3(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(join(folder_nn3, 'bandedges_1d.dat'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords['position'].name, 'position')
        self.assertEqual(df.coords['position'].unit, 'nm')
        self.assertEqual(df.coords['position'].value.size, 258)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertEqual(df.variables['Gamma_bandedge'].name, 'Gamma_bandedge')
        self.assertEqual(df.variables['Gamma_bandedge'].unit, 'eV')
        self.assertEqual(df.variables['Gamma_bandedge'].value.size, 258)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

        df = outputs.DataFile(join(folder_nn3, 'wf_shift_1d.dat'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 10)
        self.assertEqual(df.coords['position'].name, 'position')
        self.assertEqual(df.coords['position'].unit, 'nm')
        self.assertEqual(df.coords['position'].value.size, 121)
        self.assertEqual(df.variables['psi^2_1'].name, 'psi^2_1')
        self.assertEqual(df.variables['psi^2_1'].unit, default_unit)
        self.assertEqual(df.variables['psi^2_1'].value.size, 121)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

        df = outputs.DataFile(join(folder_nn3, 'bandedges_2d_cut.dat'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords['position'].name, 'position')
        self.assertEqual(df.coords['position'].unit, 'nm')
        self.assertEqual(df.coords['position'].value.size, 258)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertEqual(df.variables['Gamma_bandedge'].name, 'Gamma_bandedge')
        self.assertEqual(df.variables['Gamma_bandedge'].unit, 'eV')
        self.assertEqual(df.variables['Gamma_bandedge'].value.size, 258)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

    def test_dat_FirstVarIsCoord(self):
        df = outputs.DataFile(join(folder_nn3, 'bandedges_1d.dat'), product='nextnano3', FirstVarIsCoordFlag = False)
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 7)
        self.assertEqual(df.variables['position'].name, 'position')
        self.assertEqual(df.variables['position'].unit, 'nm')
        self.assertEqual(df.variables['position'].value.size, 258)
        self.assertEqual(df.variables['Gamma_bandedge'].name, 'Gamma_bandedge')
        self.assertEqual(df.variables['Gamma_bandedge'].unit, 'eV')
        self.assertEqual(df.variables['Gamma_bandedge'].value.size, 258)
        self.assertEqual(df.metadata['ndim'], 0)
        self.assertEqual(df.metadata['dkeys'], [])

        df = outputs.DataFile(join(folder_nn3, 'wf_shift_1d.dat'), product='nextnano3', FirstVarIsCoordFlag = True)
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 10)
        self.assertEqual(df.coords['position'].name, 'position')
        self.assertEqual(df.coords['position'].unit, 'nm')
        self.assertEqual(df.coords['position'].value.size, 121)
        self.assertEqual(df.variables['psi^2_1'].name, 'psi^2_1')
        self.assertEqual(df.variables['psi^2_1'].unit, default_unit)
        self.assertEqual(df.variables['psi^2_1'].value.size, 121)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])


    def test_avs1D(self):
        df = outputs.DataFile(join(folder_nn3, 'cb_Gamma_avs.fld'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(list(df.coords.keys()), ['x'])
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 105)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['cb_Gamma'].name, 'cb_Gamma')
        self.assertEqual(df.variables['cb_Gamma'].unit, 'eV')
        self.assertEqual(df.variables['cb_Gamma'].value.shape, (105,))
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dims'], [105])
        self.assertEqual(df.metadata['field'], 'rectilinear')

    def test_avs(self):
        df = outputs.DataFile(join(folder_nn3, 'bandedges_2d.fld'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 258)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, 'nm')
        self.assertEqual(df.coords['y'].value.size, 11)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertEqual(df.variables['Gamma_bandedge'].name, 'Gamma_bandedge')
        self.assertEqual(df.variables['Gamma_bandedge'].unit, 'eV')
        self.assertEqual(df.variables['Gamma_bandedge'].value.shape, (258, 11))
        self.assertEqual(df.metadata['ndim'], 2)
        self.assertEqual(df.metadata['dims'], [258, 11])
        self.assertEqual(df.metadata['field'], 'rectilinear')

    def test_avs2D(self):
        df = outputs.DataFile(join(folder_nn3, '2Dcb1_sg1_deg1_psi_ev001.fld'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 61)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, 'nm')
        self.assertEqual(df.coords['y'].value.size, 61)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(df.variables[0], df.variables['psi_real'])
        self.assertEqual(df.variables[0].name, 'psi_real')
        self.assertEqual(df.variables[0].unit, 'nm^-1')
        self.assertEqual(df.variables[0].value.shape, (61, 61))
        self.assertEqual(df.variables[1], df.variables['psi_imag'])
        self.assertEqual(df.variables[1].name, 'psi_imag')
        self.assertEqual(df.variables[1].unit, 'nm^-1')
        self.assertEqual(df.variables[1].value.shape, (61, 61))
        self.assertEqual(df.metadata['ndim'], 2)
        self.assertEqual(df.metadata['dims'], [61, 61])
        self.assertEqual(df.metadata['field'], 'rectilinear')

    def test_avs3D(self):
        df = outputs.DataFile(join(folder_nn3, '3Dcb1_sg1_deg1_psi_squared_ev001.fld'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ['x', 'y', 'z'])
        self.assertEqual(df.coords['x'].unit, 'nm')
        self.assertEqual(df.coords['x'].value.size, 19)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, 'nm')
        self.assertEqual(df.coords['y'].value.size, 19)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(df.coords['z'].unit, 'nm')
        self.assertEqual(df.coords['z'].value.size, 9)
        self.assertEqual(df.coords['z'].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['psi_squared'].name, 'psi_squared')
        self.assertEqual(df.variables['psi_squared'].unit, 'nm^-3')
        self.assertEqual(df.variables[0].value.shape, (19, 19, 9))
        self.assertEqual(df.metadata['ndim'], 3)
        self.assertEqual(df.metadata['dims'], [19, 19, 9])
        self.assertEqual(df.metadata['field'], 'rectilinear')

    def test_vtr_LocalDOS(self):
        df = outputs.DataFile(join(folder_nn3, 'LocalDOS_sg1_deg1_Lead1.vtr'), product='nextnano3')

        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 51)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 101)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['LocalDOS_sg1_deg1_Lead1'].name, 'LocalDOS_sg1_deg1_Lead1')
        self.assertEqual(df.variables['LocalDOS_sg1_deg1_Lead1'].unit, 'nm^-1eV^-1')
        self.assertEqual(df.variables['LocalDOS_sg1_deg1_Lead1'].value.shape, (51, 101))
        self.assertEqual(df.metadata, {})

    def test_vtr_2Dcb1(self):
        df = outputs.DataFile(join(folder_nn3, '2Dcb1_sg1_deg1_psi_ev001.vtr'), product='nextnano3')

        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 61)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 61)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(len(df.variables.keys()), 2)
        self.assertEqual(df.variables[0], df.variables['psi_real'])
        self.assertEqual(df.variables[0].name, 'psi_real')
        self.assertEqual(df.variables[0].unit, 'nm^-1')
        self.assertEqual(df.variables[0].value.shape, (61, 61))
        self.assertEqual(df.variables[1], df.variables['psi_imag'])
        self.assertEqual(df.variables[1].name, 'psi_imag')
        self.assertEqual(df.variables[1].unit, 'nm^-1')
        self.assertEqual(df.variables[1].value.shape, (61, 61))
        self.assertEqual(df.metadata, {})

    def test_vtr_3Dcb1(self):
        df = outputs.DataFile(join(folder_nn3, '3Dcb1_sg1_deg1_psi_squared_ev001.vtr'), product='nextnano3')

        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ['x', 'y', 'z'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 19)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 19)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(df.coords['z'].unit, default_unit)
        self.assertEqual(df.coords['z'].value.size, 9)
        self.assertEqual(df.coords['z'].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['psi_squared'].name, 'psi_squared')
        self.assertEqual(df.variables['psi_squared'].unit, 'nm^-3')
        self.assertEqual(df.variables['psi_squared'].value.shape, (19, 19, 9))
        self.assertEqual(df.metadata, {})

    def test_txt(self):
        df = outputs.DataFile(join(folder_nn3, 'variables_input.txt'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 3)
        self.assertEqual(df.variables['var_1'].name, 'var_1')
        self.assertEqual(df.variables['var_1'].unit, default_unit)
        self.assertEqual(df.variables['var_1'].value, -47.5)
        self.assertEqual(df.variables['var_2'].value, 75)
        self.assertEqual(df.variables['string'].value, '"hello"')

        df = outputs.DataFile(join(folder_nn3, 'variables_database.txt'), product='nextnano3')
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 0)

    def test_rest(self):
        files = ['example.log', 'example.in']
        for file in files:
            self.assertRaises(NotImplementedError, outputs.DataFile, join(folder_nnp, file), 'nextnano3')


    # TODO add tests for loading binary (different binary + avs not one file)

class TestOutputs_negf(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(join(folder_negf, 'ReducedRealSpaceModes.dat'), product='nextnano.NEGF')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords['Position'], df.coords[0])
        self.assertEqual(df.coords['Position'].name, 'Position')
        self.assertEqual(df.coords['Position'].unit, 'nm')
        self.assertEqual(df.coords['Position'].value.size, 608)
        self.assertEqual(len(df.variables.keys()), 19)
        self.assertEqual(df.variables['Conduction_BandEdge'], df.variables[0])
        self.assertEqual(df.variables['Conduction_BandEdge'].name, 'Conduction_BandEdge')
        self.assertEqual(df.variables['Conduction_BandEdge'].unit, 'eV')
        self.assertEqual(df.variables['Conduction_BandEdge'].value.size, 608)
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'], df.variables[1])
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'].name, 'Psi_1 (lev.1 per.0)')
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'].unit, 'a.u.')
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'].value.size, 608)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

        df = outputs.DataFile(join(folder_negf, 'E_p (Kane energy).dat'), product='nextnano.NEGF')
        self.assertEqual(df.coords['Position'], df.coords[0])
        self.assertEqual(df.coords['Position'].name, 'Position')
        self.assertEqual(df.coords['Position'].unit, 'nm')
        self.assertEqual(df.coords['Position'].value.size, 203)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['E_p'].name, 'E_p')
        self.assertEqual(df.variables['E_p'].unit, default_unit)
        self.assertEqual(df.variables['E_p'].value.size, 203)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])


    def test_dat_FirstVarIsCoordFlag(self):
        df = outputs.DataFile(join(folder_negf, 'ReducedRealSpaceModes.dat'), product='nextnano.NEGF', FirstVarIsCoordFlag = False)
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 20)
        self.assertEqual(df.variables['Position'].name, 'Position')
        self.assertEqual(df.variables['Position'].unit, 'nm')
        self.assertEqual(df.variables['Position'].value.size, 608)
        self.assertEqual(df.variables['Conduction_BandEdge'], df.variables[1])
        self.assertEqual(df.variables['Conduction_BandEdge'].name, 'Conduction_BandEdge')
        self.assertEqual(df.variables['Conduction_BandEdge'].unit, 'eV')
        self.assertEqual(df.variables['Conduction_BandEdge'].value.size, 608)
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'], df.variables[2])
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'].name, 'Psi_1 (lev.1 per.0)')
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'].unit, 'a.u.')
        self.assertEqual(df.variables['Psi_1 (lev.1 per.0)'].value.size, 608)
        self.assertEqual(df.metadata['ndim'], 0)
        self.assertEqual(df.metadata['dkeys'], [])

        df = outputs.DataFile(join(folder_negf, 'E_p (Kane energy).dat'), product='nextnano.NEGF', FirstVarIsCoordFlag=True)
        self.assertEqual(df.coords['Position'], df.coords[0])
        self.assertEqual(df.coords['Position'].name, 'Position')
        self.assertEqual(df.coords['Position'].unit, 'nm')
        self.assertEqual(df.coords['Position'].value.size, 203)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['E_p'].name, 'E_p')
        self.assertEqual(df.variables['E_p'].unit, default_unit)
        self.assertEqual(df.variables['E_p'].value.size, 203)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

    def test_vtr(self):
        df = outputs.DataFile(join(folder_negf, 'CarrierDensity_energy_resolved.vtr'), product='nextnano.NEGF')
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 176)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 82)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['Carrier density'].name, 'Carrier density')
        self.assertEqual(df.variables['Carrier density'].unit, 'cm^-3 eV^-1')
        self.assertEqual(df.variables['Carrier density'].value.shape, (176, 82))
        self.assertEqual(df.metadata, {})

    def test_rest(self):
        files = ['example.log', 'example.in', 'Material_Database.xml', 'Convergence.txt']
        for file in files:
            self.assertRaises(NotImplementedError, outputs.DataFile, join(folder_negf, file), 'nextnano.NEGF')

# NEGF C# and NEGF C++ have identical outputs. Maybe no need to test.

class TestOutputs_msb(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(join(folder_msb, 'BandEdge_conduction.dat'), product='nextnano.MSB')
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords['Position'].name, 'Position')
        self.assertEqual(df.coords['Position'].unit, 'nm')
        self.assertEqual(df.coords['Position'].value.size, 100)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['Conduction Band Edge'].name, 'Conduction Band Edge')
        self.assertEqual(df.variables['Conduction Band Edge'].unit, 'eV')
        self.assertEqual(df.variables['Conduction Band Edge'].value.size, 100)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

    def test_dat_FirstVarIsCoordFlag(self):
        df = outputs.DataFile(join(folder_msb, 'BandEdge_conduction.dat'), product='nextnano.MSB', FirstVarIsCoordFlag = False)
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 2)
        self.assertEqual(df.variables['Position'].name, 'Position')
        self.assertEqual(df.variables['Position'].unit, 'nm')
        self.assertEqual(df.variables['Position'].value.size, 100)
        self.assertEqual(df.variables['Conduction Band Edge'].name, 'Conduction Band Edge')
        self.assertEqual(df.variables['Conduction Band Edge'].unit, 'eV')
        self.assertEqual(df.variables['Conduction Band Edge'].value.size, 100)
        self.assertEqual(df.metadata['ndim'], 0)
        self.assertEqual(df.metadata['dkeys'], [])

        df = outputs.DataFile(join(folder_msb, 'BandEdge_conduction.dat'), product='nextnano.MSB', FirstVarIsCoordFlag = True)
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords['Position'].name, 'Position')
        self.assertEqual(df.coords['Position'].unit, 'nm')
        self.assertEqual(df.coords['Position'].value.size, 100)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables['Conduction Band Edge'].name, 'Conduction Band Edge')
        self.assertEqual(df.variables['Conduction Band Edge'].unit, 'eV')
        self.assertEqual(df.variables['Conduction Band Edge'].value.size, 100)
        self.assertEqual(df.metadata['ndim'], 1)
        self.assertEqual(df.metadata['dkeys'], [0])

    def test_vtr(self):
        df = outputs.DataFile(join(folder_msb, 'DOS_Lead_Source_position_resolved.vtr'), product='nextnano.MSB')
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ['x', 'y'])
        self.assertEqual(df.coords['x'].unit, default_unit)
        self.assertEqual(df.coords['x'].value.size, 50)
        self.assertEqual(df.coords['x'].dim, 0)
        self.assertEqual(df.coords['y'].unit, default_unit)
        self.assertEqual(df.coords['y'].value.size, 401)
        self.assertEqual(df.coords['y'].dim, 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables[0], df.variables['Density of States (Source)'])
        self.assertEqual(df.variables[0].name, 'Density of States (Source)')
        self.assertEqual(df.variables[0].unit, '1/eV/nm')
        self.assertEqual(df.variables[0].value.shape, (50, 401))
        self.assertEqual(df.metadata, {})


class TestOutput(unittest.TestCase):

    def test_data(self):
        df = outputs.DataFile(join(folder_nnp, 'bandedges_2d.fld'), product='nextnano++')
        self.assertEqual(len(df.data), len(df.coords) + len(df.variables))
        for key, value in df.coords.items():
            self.assertEqual(df.data[key], value)
        for key, value in df.variables.items():
            self.assertEqual(df.data[key], value)

    def test_access_by_index(self):
        df = outputs.DataFile(join(folder_nnp, 'bandedges_2d.fld'), product='nextnano++')
        for key, value in df.coords.items():
            self.assertEqual(df[key], value)
        for key, value in df.variables.items():
            self.assertEqual(df[key], value)

    def test_for_loop(self):
        df = outputs.DataFile(join(folder_nnp, 'bandedges_2d.fld'), product='nextnano++')
        for i, dfi in enumerate(df):
            self.assertEqual(df.data[i], dfi)

class TestDataFolder(unittest.TestCase):
    def test_init(self):
        dummy_folder = os.path.join('tests','dummy')
        self.assertRaises(ValueError, outputs.DataFolder, dummy_folder)

        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)
        self.assertTrue('datafiles' in datafolder.folders)
        self.assertTrue(os.path.join(tests_folder, 'test_mycollections.py') in datafolder.files)
        self.assertTrue(os.path.join(tests_folder, 'test_outputs.py') in datafolder.files)
        self.assertFalse('NotExistedFile.py' in datafolder.files)
        self.assertFalse(os.path.join(tests_folder,'NotExistedFile.py') in datafolder.files)

        datafolder  = outputs.DataFolder(folder_nnp)

        self.assertTrue(datafolder.folders)
        self.assertTrue(datafolder.files)

        self.assertEqual(len(datafolder.files),17)
        self.assertTrue(os.path.join(folder_nnp,'bandedges_2d_old.fld') in datafolder.files)


    def test_navigation(self):
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)
        self.assertTrue('configs' in datafolder.__dict__)
        self.assertIsInstance(datafolder.configs, outputs.DataFolder)
        self.assertIsInstance(datafolder.datafiles.nextnano3, outputs.DataFolder)
        self.assertIsInstance(datafolder.datafiles.folders['nextnano++'], outputs.DataFolder)
        nnp_folder = datafolder.datafiles.folders['nextnano++']
        self.assertTrue(os.path.samefile(nnp_folder.fullpath, folder_nnp))
        self.assertTrue(os.path.join(folder_nnp, 'bandedges_2d_old.fld') in nnp_folder.files)


    def test_filenamess(self):

        datafolder  = outputs.DataFolder(folder_nnp)
        self.assertIn('only_variables.in',datafolder.filenames())
        self.assertNotIn('only_variables_0.in',datafolder.filenames())
        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'),datafolder.filenames())
        self.assertEqual(len(datafolder.filenames()), 17)
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)

        self.assertIn('test_all.py',datafolder.filenames())
        self.assertIn('test_outputs.py',datafolder.filenames())

        self.assertNotIn('only_variables.in',datafolder.filenames())
        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'),datafolder.filenames())

    def test_find(self):
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)
        self.assertEqual(len(datafolder.find('')),12)

        self.assertIn(os.path.join(tests_folder,'__init__.py'),datafolder.find(''))
        self.assertIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find(''))


        self.assertNotIn(os.path.join(folder_nnp,'only_variables.in'),datafolder.find(''))
        self.assertNotIn('only_variables.in',datafolder.find(''))



        self.assertNotIn(os.path.join(tests_folder, '__init__.py'), datafolder.find('test'))
        self.assertIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('test'))

        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('test'))
        self.assertNotIn('only_variables.in', datafolder.find('test'))
        self.assertNotIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('test_i'))
        self.assertEqual(len(datafolder.find('.vtr')), 0)

        datafolder = outputs.DataFolder(folder_nnp)
        self.assertNotIn(os.path.join(tests_folder, '__init__.py'), datafolder.find('tests'))
        self.assertNotIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('tests'))

        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('tests'))
        self.assertIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('only'))
        self.assertEqual(len(datafolder.find('bandedges')), 7)
        self.assertNotIn('only_variables.in', datafolder.find('abls'))
        self.assertEqual(len(datafolder.find('.vtr')),2)



        #deep
        datafolder = outputs.DataFolder(tests_folder)
        self.assertIn(os.path.join(tests_folder, '__init__.py'), datafolder.find('',deep = True))
        self.assertIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('',deep = True))
        self.assertIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('',deep = True))
        self.assertNotIn('only_variables.in', datafolder.find('',deep = True))

        self.assertNotIn(os.path.join(tests_folder, '__init__.py'), datafolder.find('test', deep = True))
        self.assertIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('test', deep = True))

        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('test', deep = True))
        self.assertIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('only', deep=True))
        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('nextnano', deep=True))
        self.assertNotIn('only_variables.in', datafolder.find('only'))
        self.assertNotIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('test_i', deep = True))
        self.assertNotEqual(len(datafolder.find('.vtr', deep = True)), 0)


        datafolder = outputs.DataFolder(folder_nnp)

        self.assertNotIn(os.path.join(tests_folder, '__init__.py'), datafolder.find('', deep=True))
        self.assertNotIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('', deep=True))
        self.assertIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('', deep=True))
        self.assertNotIn('only_variables.in', datafolder.find('', deep=True))

        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('test', deep=True))
        self.assertIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('only', deep=True))
        self.assertNotIn(os.path.join(folder_nnp, 'only_variables.in'), datafolder.find('nextnano', deep=True))
        self.assertNotIn('only_variables.in', datafolder.find('only'))
        self.assertNotIn(os.path.join(tests_folder, 'test_commands.py'), datafolder.find('test_i', deep=True))
        self.assertNotEqual(len(datafolder.find('.vtr', deep=True)), 0)
        self.assertEqual(len(datafolder.find('bandedges', deep = True)), 17)

    def test_find_multiple(self):
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)
        self.assertEqual(len(datafolder.find_multiple(('',))),12)

        self.assertIn(os.path.join(tests_folder,'__init__.py'),datafolder.find_multiple(('',)))


        self.assertNotIn('only_variables.in',datafolder.find_multiple(('',)))


        datafolder = outputs.DataFolder(folder_nnp)
        self.assertNotIn(os.path.join(tests_folder, '__init__.py'), datafolder.find_multiple(('tests',)))

        self.assertEqual(len(datafolder.find_multiple(('bandedges',))), 7)
        self.assertEqual(len(datafolder.find_multiple(('bandedges', '2d'))), 5)

        self.assertNotIn('only_variables.in', datafolder.find_multiple(('abls',)))

        self.assertEqual(len(datafolder.find_multiple(('.vtr',))), 2)
        self.assertEqual(len(datafolder.find_multiple(('.vt', 'r'))), 2)

    def test_go_to(self):
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)
        self.assertIsInstance(datafolder.go_to('datafiles'),outputs.DataFolder)
        self.assertIsInstance(datafolder.go_to('__init__.py'), str)
        self.assertTrue(os.path.samefile(datafolder.go_to('__init__.py'),os.path.join(tests_folder,'__init__.py')))
        self.assertFalse(os.path.samefile(datafolder.go_to('__init__.py'), os.path.join(tests_folder)))

        datafolder_nnp = datafolder.go_to('datafiles','nextnano++')

        self.assertTrue(os.path.samefile(datafolder_nnp.fullpath,folder_nnp))
        self.assertEqual(len(datafolder_nnp.files), 17)

    def test_file(self):
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)
        self.assertRaises(ValueError, datafolder.file, 'hello_world')

        self.assertEqual(datafolder.datafiles.folders['nextnano++'].file('bandedges.vtr'),os.path.join(folder_nnp,'bandedges.vtr'))

        self.assertRaises(ValueError,datafolder.datafiles.folders['nextnano++'].file, 'bandedgesxxxx.vtr')
        warnings.filterwarnings('ignore', 'More than one file match')
        self.assertTrue(datafolder.datafiles.folders['nextnano++'].file('bandedges'))
        warnings.filterwarnings('default')


    def test_make_tree(self):
        tests_folder = 'tests'
        datafolder = outputs.DataFolder(tests_folder)

        tree_list = datafolder.make_tree()

        self.assertEqual(tree_list[1],'    configs/')
        self.assertEqual(tree_list[2],'        .nnconfig')
        self.assertEqual(tree_list[4],'        nextnano++/')

















if __name__ == '__main__':
    unittest.main()



from nextnanopy.utils.mycollections import DictList
from nextnanopy.inputs import InputFileTemplate
from nextnanopy.negf.defaults import is_negf_input_text, InputVariable_NEGF, parse_negf_variable_name
from nextnanopy.utils.formatting import autofmt_variable_value
import xml.etree.ElementTree as ET

class InputFile(InputFileTemplate):
    def load_variables(self):
        #parser = ET.XMLParser(target=ET.TreeBuilder(insert_comments=True)) #Does not supported in python3.7
        parser = ET.XMLParser()
        root = ET.fromstring(self.text, parser = parser)
        varsection = root.find('Variables')
        variables = DictList()
        if not varsection:
            self.variables = variables
            return self.variables
        vars = varsection.findall('Constant')


        for var in vars:
            name_element = var.find('Name')
            name = parse_negf_variable_name(name_element.text)
            comment = name_element.get('Comment')
            value_element = var.find('Value')
            value = autofmt_variable_value(value_element.text)
            unit = value_element.get('Unit')
            var = InputVariable_NEGF(name=name, value=value, unit = unit,  comment=comment)
            variables[name] = var
        self.variables = variables
        return self.variables

    @property
    def lines(self):
        root = ET.fromstring(self.raw_text)
        varsection = root.find('Variables')
        variables = self.variables
        if varsection:
            vars = varsection.findall('Constant')
            for var_element in vars:
                for variable in variables:
                    name_element = var_element.find('Name')
                    if name_element.text ==  '$'+variable.name:
                        value_element = var_element.find('Value')
                        value_element.text = str(variable.value)
                        name_element.set('Comment', variable.comment)
                        value_element.set('Unit', variable.unit)


        xmlstr = ET.tostring(root, encoding='unicode', method='xml')
        lines_result = xmlstr.splitlines()
        return lines_result






    def validate(self):
        if not is_negf_input_text(self.raw_text):
            raise ValueError(f'Not valid nextnano.NEGF input file')



from nextnanopy.outputs import DataFile, DataFolder
from nextnanopy.commands import execute
from nextnanopy.inputs import InputFile, Sweep
from nextnanopy.defaults import NNConfig
# from nextnanopy.shapes import GdsPolygonsRaw

config = NNConfig()





from copy import deepcopy
import numpy as np

default_unit = 'a.u.'


class Data(object):
    """
    This class stores any kind of information from nextnano files (input files, data files).
    This is a template class for datasets like Variables, Coords, InputVariables, etc.

    ...

    Parameters
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    unit : str, optional
        unit of the value (default is None)
    metadata : dict, optional
        extra information (default is {})
    label_fmt : method, optional
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'

    Attributes
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    unit : str
        unit of the value (default is None)
    metadata : dict
        extra information (default is {})
    label_fmt : method
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'
    """

    params = ['name', 'value', 'unit', 'metadata'],

    def __init__(self, name, value, unit=None, metadata={}, label_fmt=None, *args,
                 **kwargs):
        self.name = str(name)
        self.value = np.array(value)
        if unit is None or unit == '':
            unit = default_unit
        self.unit = str(unit)
        if label_fmt is None:
            label_fmt = lambda name, unit: f'{name} ({unit})'
        self.label_fmt = label_fmt
        self.metadata = metadata

    def parameters(self):
        dict_ = {}
        for param in self.params:
            value = getattr(self, param)
            dict_[param] = value
        dict_ = deepcopy(dict_)
        return dict_

    @property
    def label(self):
        return self.label_fmt(self.name, self.unit)

    def __repr__(self):
        cname = self.__class__.__name__
        out = f'{cname} - {str(self)}'
        return out

    def __str__(self):
        return f'name: {self.name}'


class Variable(Data):
    """
    This class stores independent variables from data files.

    ...

    Parameters
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    unit : str, optional
        unit of the value (default is None)
    metadata : dict, optional
        extra information (default is {})
    label_fmt : method, optional
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'

    Attributes
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    unit : str
        unit of the value (default is None)
    metadata : dict
        extra information (default is {})
    label_fmt : method
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'


    Methods
    ----------
    get_value()
        return a copy of the value

    """
    params = ['name', 'value', 'unit', 'metadata']

    def __init__(self, name, value, unit=None, metadata={}, **kwargs):
        super().__init__(name, value, unit, metadata, **kwargs)

    def get_value(self):
        value = deepcopy(self.value)
        return value

    def __str__(self):
        return f'name: {self.name} - unit: {self.unit} - shape: {self.value.shape}'


class Coord(Data):
    """
    This class stores the coordinates from data files.

    ...

    Parameters
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    dim : int
        dim for the Variable.value
        For example, for a dataset of (100, 2) and a coordinate x of 100 pts, dim = 0
    unit : str, optional
        unit of the value (default is None)
    offset : not defined, optional
        offset to be added to value (default is 0)
    metadata : dict, optional
        extra information (default is {})
    label_fmt : method, optional
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'

    Attributes
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    dim : int
        dim for the Variable.value
        For example, for a dataset of (100, 2) and a coordinate x of 100 pts, dim = 0
    unit : str
        unit of the value (default is None)
    offset : not defined
        offset to be added to value (default is 0)
    metadata : dict
        extra information (default is {})
    label_fmt : method
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'
    valueo : not defined
        value with offset

    Methods
    ----------
    get_value(use_offset=False)
        return a copy of the value with or without the offset
    """

    params = ['name', 'value', 'unit', 'offset', 'dim', 'metadata']

    def __init__(self, name, value, dim, unit=None, offset=0, metadata={}, **kwargs):
        super().__init__(name, value, unit, metadata, **kwargs)
        self.dim = int(dim)
        self.offset = np.array(offset)
        self.valueo = self.get_value(use_offset=True)

    def get_value(self, use_offset=False):
        value = deepcopy(self.value)
        if use_offset:
            value += self.offset
        return value

    def __str__(self):
        return f'name: {self.name} - unit: {self.unit} - shape: {self.value.shape} - dim: {self.dim}'


class InputVariable(Data):
    """
    Template class for the input variables. For each nextnano product, it will be specified
    the variable character (e.g $ for nextnano++) and the comment character (e.g # for nextnano++).

    ...

    Parameters
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    unit : str, optional
        unit of the value (default is '')
    comment : str, optional
        (default is '')
    metadata : dict, optional
        extra information (default is {})
    label_fmt : method, optional
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'

    Attributes
    ----------
    name : str
        name of the dataset
    value : not defined
        stored value
    unit : str
        unit of the value (default is '')
    comment : str
        (default is '')
    metadata : dict
        extra information (default is {})
    label_fmt : method
        formatting label with label_fmt(name, unit) (default is None)
        If it is None, label_fmt = lambda name, unit: f'{name} ({unit})'
    text : str
        return the raw text for the input file

    Methods
    ----------
    get_value(use)
        return a copy of the value
    """

    params = ['name', 'value', 'unit', 'comment', 'metadata']
    var_char = ''
    com_char = ''

    def __init__(self, name, value, unit='', comment='', metadata={}, **kwargs):
        super().__init__(name, value, unit, metadata, **kwargs)
        self.comment = comment

    def get_value(self):
        value = deepcopy(self.value)
        return value

    @property
    def text(self):
        t = f'{self.var_char}{self.name} = {self.value}'
        if self.comment:
            t = f'{t} {self.com_char} {self.comment}'
        return t

    def __str__(self):
        return self.text



import unittest
from nextnanopy import commands
from nextnanopy.utils.formatting import _path, _bool

import os

folder_nnp = os.path.join('tests', 'datafiles', 'nextnano++')
folder_nn3 = os.path.join('tests', 'datafiles', 'nextnano3')
folder_negf = os.path.join('tests', 'datafiles', 'nextnano.NEGF')
folder_msb = os.path.join('tests', 'datafiles', 'nextnano.MSB')


class TestCommands(unittest.TestCase):

    def test_commands_nnp(self):
        self.maxDiff = None
        inputfile = os.path.join(folder_nnp, 'example.in')
        exe = os.path.join('nextnano++', 'bin 64bit', 'nextnano++_Intel_64bit.exe')
        runmode = '--resume'
        no_file_options = '--autosave --logfile'
        license = os.path.join(r'nextnanopy\License', 'License_nnp.lic')
        database = os.path.join('nextnano++', 'Syntax', 'database_nnp.in')
        outputdirectory = r'tests\datafiles'
        threads = 4
        cmd = f'"{exe}" {runmode} --license "{license}" --database "{database}" --threads {threads} --outputdirectory "{outputdirectory}" --noautooutdir {no_file_options} "{inputfile}"'
        kwargs = dict(inputfile=inputfile, runmode=runmode, exe=exe, license=license, database=database, outputdirectory=outputdirectory,
                      threads=threads, no_file_options=no_file_options)
        from nextnanopy.nnp.defaults import command_nnp
        self.assertEqual(command_nnp(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_commands_nn3(self):
        self.maxDiff = None
        inputfile = os.path.join(folder_nn3, 'example.in')
        exe = os.path.join('nextnano++', 'bin 64bit', 'nextnano++_Intel_64bit.exe')
        license = os.path.join(r'nextnanopy\License', 'License_nnp.lic')
        database = os.path.join('nextnano++', 'Syntax', 'database_nnp.in')
        outputdirectory = r'tests\datafiles'
        threads = 4
        debuglevel = 0
        cancel = -1
        softkill = -1
        no_file_options = '-log -parse'
        cmd = f'"{exe}" -license "{license}" -inputfile "{inputfile}" -database "{database}" -threads {threads} -outputdirectory "{outputdirectory}" -debuglevel {debuglevel} -cancel {cancel} -softkill {softkill} {no_file_options}'
        kwargs = dict(inputfile=inputfile, exe=exe, license=license, database=database, outputdirectory=outputdirectory,
                      threads=threads, debuglevel=debuglevel, cancel=cancel, softkill=softkill, no_file_options=no_file_options)
        from nextnanopy.nn3.defaults import command_nn3
        self.assertEqual(command_nn3(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_commands_negf(self):
        self.maxDiff = None
        inputfile = os.path.join(folder_negf, 'example.xml')
        exe = os.path.join('nextnano.NEGF', 'nextnano.NEGF.exe')
        license = os.path.join(r'License', 'License_nnQCL.lic')
        database = os.path.join('nextnano.NEGF', 'Material_Database.xml')
        outputdirectory = r'tests\datafiles'
        threads = 4
        cmd = f'"{exe}" "{inputfile}" "{outputdirectory}" "{database}" "{license}" -threads {threads}'
        kwargs = dict(inputfile=inputfile, exe=exe, license=license, database=database, outputdirectory=outputdirectory,
                      threads=threads)
        from nextnanopy.negf.defaults import command_negf
        self.assertEqual(command_negf(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_commands_msb(self):
        from nextnanopy.msb.defaults import command_msb

        self.maxDiff = None
        inputfile = os.path.join(folder_msb, 'example.xml')
        exe = os.path.join('nextnano.MSB', 'nextnano.MSB.exe')
        license = os.path.join(r'License', 'License_nnMSB.lic')
        database = os.path.join('nextnano.MSB', 'Materials.xml')
        outputdirectory = r'tests\datafiles'
        debug = 0

        cmd = f'"{exe}" -inputfile "{inputfile}" -license "{license}" -database "{database}" -outputdirectory "{outputdirectory}"'
        kwargs = dict(inputfile=inputfile, exe=exe, license=license, database=database, outputdirectory=outputdirectory,
                      debug=debug)
        self.assertEqual(command_msb(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

        debug = 1
        cmd = f'"{exe}" -inputfile "{inputfile}" -license "{license}" -database "{database}" -outputdirectory "{outputdirectory}" -debug 1'
        kwargs = dict(inputfile=inputfile, exe=exe, license=license, database=database, outputdirectory=outputdirectory,
                      debug=debug)
        self.assertEqual(command_msb(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_path(self):
        self.assertEqual(_path('aa\nb.test'), '"aa\nb.test"')
        self.assertEqual(_path('aa\nb'), '"aa\nb"')
        self.assertEqual(_path(''), '')
        self.assertEqual(_path(2), '"2"')
        self.assertEqual(_path(None), None)

    def test_execute(self):
        self.assertRaises(ValueError, commands.execute, inputfile='', exe='', license='', database='',
                          outputdirectory='')
        self.assertRaises(ValueError, commands.execute, inputfile=r'test\datafiles', exe='', license='', database='',
                          outputdirectory='')

    def test_bool(self):
        self.assertEqual(_bool(''), False)
        self.assertEqual(_bool(None), False)
        self.assertEqual(_bool('1'), True)
        self.assertEqual(_bool('0'), True)
        self.assertEqual(_bool(0), True)
        self.assertEqual(_bool(1), True)


if __name__ == '__main__':
    unittest.main()



import os
from nextnanopy.defaults import NNConfig
import unittest
import platform 

system = platform.system()


class Test_NNConfig(unittest.TestCase):
    def test_default_nn3(self):
        config = NNConfig()
        
        if system == 'Windows':
            default_fullpath = os.path.join(os.environ['HOMEDRIVE'], os.environ['HOMEPATH'], '.nextnanopy-config')
        else:
            default_fullpath = os.path.join(os.environ['HOME'], '.nextnanopy-config')

        self.assertEqual(config.fullpath, default_fullpath)
        self.assertTrue('nextnano3' in config.validators.keys())
        self.assertTrue('nextnano3' in config.defaults.keys())
        self.assertTrue('nextnano3' in config.config.keys())
        options = ['exe', 'license', 'database', 'threads', 'outputdirectory','debuglevel','cancel','softkill']
        for option in options:
            self.assertTrue(option in config.validators['nextnano3'].keys())
            self.assertTrue(option in config.defaults['nextnano3'].keys())
            self.assertTrue(option in config.config['nextnano3'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))

    def test_default_nnp(self):
        config = NNConfig()
        if system == 'Windows':
        
        	default_fullpath = os.path.join(os.environ['HOMEDRIVE'], os.environ['HOMEPATH'], '.nextnanopy-config')
        else:
        	default_fullpath = os.path.join(os.environ['HOME'], '.nextnanopy-config')

        self.assertEqual(config.fullpath, default_fullpath)
        self.assertTrue('nextnano++' in config.validators.keys())
        self.assertTrue('nextnano++' in config.defaults.keys())
        self.assertTrue('nextnano++' in config.config.keys())
        options = ['exe', 'license', 'database', 'threads', 'outputdirectory']
        for option in options:
            self.assertTrue(option in config.validators['nextnano++'].keys())
            self.assertTrue(option in config.defaults['nextnano++'].keys())
            self.assertTrue(option in config.config['nextnano++'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))

    def test_default_negf(self):
        config = NNConfig()
        if system == 'Windows':
            default_fullpath = os.path.join(os.environ['HOMEDRIVE'], os.environ['HOMEPATH'], '.nextnanopy-config')
        else:
            default_fullpath = os.path.join(os.environ['HOME'], '.nextnanopy-config')

        self.assertEqual(config.fullpath, default_fullpath)
        self.assertTrue('nextnano.NEGF' in config.validators.keys())
        self.assertTrue('nextnano.NEGF' in config.defaults.keys())
        self.assertTrue('nextnano.NEGF' in config.config.keys())
        options = ['exe', 'license', 'database', 'threads', 'outputdirectory']
        for option in options:
            self.assertTrue(option in config.validators['nextnano.NEGF'].keys())
            self.assertTrue(option in config.defaults['nextnano.NEGF'].keys())
            self.assertTrue(option in config.config['nextnano.NEGF'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))

    def test_default_nnevo(self):
        config = NNConfig()
        if system == 'Windows':
            default_fullpath = os.path.join(os.environ['HOMEDRIVE'], os.environ['HOMEPATH'], '.nextnanopy-config')
        else:
            default_fullpath = os.path.join(os.environ['HOME'], '.nextnanopy-config')

        self.assertEqual(config.fullpath, default_fullpath)
        self.assertTrue('nextnanoevo' in config.validators.keys())
        self.assertTrue('nextnanoevo' in config.defaults.keys())
        self.assertTrue('nextnanoevo' in config.config.keys())
        options = ['license']
        for option in options:
            self.assertTrue(option in config.validators['nextnanoevo'].keys())
            self.assertTrue(option in config.defaults['nextnanoevo'].keys())
            self.assertTrue(option in config.config['nextnanoevo'].keys())
        option = 'exe' # no exe for nnevo
        self.assertFalse(option in config.validators['nextnanoevo'].keys())
        self.assertFalse(option in config.defaults['nextnanoevo'].keys())
        self.assertFalse(option in config.config['nextnanoevo'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))

    def test_load_nn3(self):
        fullpath = os.path.join('tests', '.nextnanopy-config')
        config = NNConfig(fullpath)

        self.assertEqual(config.fullpath, fullpath)
        self.assertTrue('nextnano3' in config.validators.keys())
        self.assertTrue('nextnano3' in config.defaults.keys())
        self.assertTrue('nextnano3' in config.config.keys())
        options = ['exe', 'license', 'database', 'threads', 'outputdirectory','debuglevel','cancel','softkill']
        for option in options:
            self.assertTrue(option in config.validators['nextnano3'].keys())
            self.assertTrue(option in config.defaults['nextnano3'].keys())
            self.assertTrue(option in config.config['nextnano3'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)

        self.assertEqual(config.config['nextnano3']['exe'], '')
        config.set('nextnano3', 'exe', 'some_path')
        self.assertEqual(config.config['nextnano3']['exe'], 'some_path')
        config.to_default()
        self.assertEqual(config.config['nextnano3']['exe'], '')
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)
        fullpath_new = os.path.join('tests', 'test.nnconfig')
        self.assertFalse(os.path.isfile(fullpath_new))
        config.save(fullpath_new)
        self.assertTrue(os.path.isfile(fullpath_new))
        self.assertEqual(config.fullpath, fullpath_new)
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)

    def test_load_nnp(self):
        fullpath = os.path.join('tests', '.nextnanopy-config')
        config = NNConfig(fullpath)

        self.assertEqual(config.fullpath, fullpath)
        self.assertTrue('nextnano++' in config.validators.keys())
        self.assertTrue('nextnano++' in config.defaults.keys())
        self.assertTrue('nextnano++' in config.config.keys())
        options = ['exe', 'license', 'database', 'threads', 'outputdirectory']
        for option in options:
            self.assertTrue(option in config.validators['nextnano++'].keys())
            self.assertTrue(option in config.defaults['nextnano++'].keys())
            self.assertTrue(option in config.config['nextnano++'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)

        self.assertEqual(config.config['nextnano++']['exe'], '')
        config.set('nextnano++', 'exe', 'some_path')
        self.assertEqual(config.config['nextnano++']['exe'], 'some_path')
        config.to_default()
        self.assertEqual(config.config['nextnano++']['exe'], '')
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)
        fullpath_new = os.path.join('tests', 'test.nnconfig')
        self.assertFalse(os.path.isfile(fullpath_new))
        config.save(fullpath_new)
        self.assertTrue(os.path.isfile(fullpath_new))
        self.assertEqual(config.fullpath, fullpath_new)
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)

    def test_load_negf(self):
        fullpath = os.path.join('tests', '.nextnanopy-config')
        config = NNConfig(fullpath)

        self.assertEqual(config.fullpath, fullpath)
        self.assertTrue('nextnano.NEGF' in config.validators.keys())
        self.assertTrue('nextnano.NEGF' in config.defaults.keys())
        self.assertTrue('nextnano.NEGF' in config.config.keys())
        options = ['exe', 'license', 'database', 'threads', 'outputdirectory']
        for option in options:
            self.assertTrue(option in config.validators['nextnano.NEGF'].keys())
            self.assertTrue(option in config.defaults['nextnano.NEGF'].keys())
            self.assertTrue(option in config.config['nextnano.NEGF'].keys())
        self.assertTrue(os.path.isfile(config.fullpath))
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)

        self.assertEqual(config.config['nextnano.NEGF']['exe'], '')
        config.set('nextnano.NEGF', 'exe', 'some_path')
        self.assertEqual(config.config['nextnano.NEGF']['exe'], 'some_path')
        config.to_default()
        self.assertEqual(config.config['nextnano.NEGF']['exe'], '')
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)
        fullpath_new = os.path.join('tests', 'test.nnconfig')
        self.assertFalse(os.path.isfile(fullpath_new))
        config.save(fullpath_new)
        self.assertTrue(os.path.isfile(fullpath_new))
        self.assertEqual(config.fullpath, fullpath_new)
        if os.path.isfile(config.fullpath):
            os.remove(config.fullpath)

if __name__ == '__main__':
    unittest.main()



import unittest
from tests.test_commands import *
from tests.test_inputs import *
from tests.test_misc import *
from tests.test_outputs import *
from tests.test_shapes import *
from tests.test_defaults import *
from tests.test_config import *
from tests.test_formatting import *
from tests.test_datasets import *
from tests.test_mycollections import *

if __name__ == '__main__':
    unittest.main()



import time


def timer(method, fmt=lambda name, t: f'[Timer] {name}: {t * 1000:2.2f} ms', apply=True):
    if not apply:
        return method

    def timed(*args, **kwargs):
        start = time.time()
        result = method(*args, **kwargs)
        stop = time.time()
        elapse = (stop - start)
        print(fmt(method.__name__, elapse))
        return result

    return timed



import numpy as np
from nextnanopy.utils.mycollections import DictList
from nextnanopy.outputs import Output, AvsAscii, Vtk, DataFileTemplate
from nextnanopy.utils.datasets import Variable, Coord
import re


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath, product='nextnano.MSB')
        self.load()

    def get_loader(self):
        if self.extension in ['.v', '.fld', '.coord']:
            loader = AvsAscii
        elif self.extension == '.vtr':
            loader = Vtk
        elif self.extension == '.txt':
            raise NotImplementedError(f'Loading nextnano.MSB datafiles with extension.txt is not implemented yet')
        elif self.extension == '.dat':
            loader = Dat
        else:
            raise NotImplementedError(f'Loading datafile with extension {self.extension} is not implemented yet')
        return loader


class Dat(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load(**loader_kwargs)

    def load(self, **loader_kwargs):
        self.load_metadata(**loader_kwargs)
        self.load_data()

    def _get_headers(self):
        headers = []
        with open(self.fullpath, 'r') as f:
            for line in f:
                try:
                    float(line.split()[0])
                    break
                except:
                    headers.append(line)
        return headers

    def _get_nb_columns(self):
        meta = self.metadata
        with open(self.fullpath, 'r') as f:
            for i, line in enumerate(f):
                if i < meta['skip_rows']:
                    continue
                line = line.replace('\n', '').strip().split()
                arr = np.array(line)
                break
        return arr.size

    def load_metadata(self, FirstVarIsCoordFlag = True):
        headers = self._get_headers()
        self.metadata['headers'] = headers
        self.metadata['skip_rows'] = len(headers)
        nb = self._get_nb_columns()
        self.metadata['nb_columns'] = nb

        if len(headers) == 0:
            raise NotImplementedError('.dat file without header')
        else:
            header = headers[-1]  # take the last one by default
        header = re.split(r'\[|\]', header)
        header = [hi.strip() for hi in header]
        columns, units = header[0::2], header[1::2]
        diff = len(columns) - len(units)
        if len(columns) > nb:
            columns = columns[0:nb]
        elif diff > 0:
            empty = [''] * diff
            units.extend(empty)
        ndim = 0
        dkeys = []
        # FirstVarIsCoordFlag = True
        for i, (column, unit) in enumerate(zip(columns, units)):
            self.metadata[i] = {'name': column, 'unit': unit}
            if FirstVarIsCoordFlag:
                ndim += 1
                dkeys.append(i)
                FirstVarIsCoordFlag = False
        self.metadata['ndim'] = ndim
        self.metadata['dkeys'] = dkeys
        return self.metadata

    def load_data(self):
        data = []
        meta = self.metadata
        with open(self.fullpath, 'r') as f:
            for i, line in enumerate(f):
                if i < meta['skip_rows']:
                    continue
                line = line.replace('\n', '').strip().split()
                if line:
                    data.append(line)
        data = np.array(data, dtype=float).T  # columns 1st index
        coords, variables = DictList(), DictList()
        dims = []
        for i, values in enumerate(data):
            vm = meta[i]
            if i in meta['dkeys']:
                #values = np.unique(values)
                dims.append(values.size)
                var = Coord(name=vm['name'], unit=vm['unit'], dim=i, value=values)
                coords[var.name] = var
            else:
                if dims:
                    values = values.reshape(*dims)
                var = Variable(name=vm['name'], unit=vm['unit'], value=values)
                variables[var.name] = var
        self.coords = coords
        self.variables = variables
        return coords, variables



import unittest
from nextnanopy.utils import mycollections
import os
from nextnanopy import defaults
from nextnanopy.nnp import defaults as nnp_defaults
from nextnanopy.nn3 import defaults as nn3_defaults
from nextnanopy.negf import defaults as negf_defaults
from nextnanopy.msb import defaults as msb_defaults


folder_nnp = os.path.join('tests', 'datafiles', 'nextnano++')
folder_nn3 = os.path.join('tests', 'datafiles', 'nextnano3')
folder_negf = os.path.join('tests', 'datafiles', 'nextnano.NEGF')
folder_negfpp = os.path.join('tests', 'datafiles', 'nextnano.NEGF++')


class TestDictlist(unittest.TestCase):
    def test_dictlist(self):
        a = mycollections.DictList(a=2, b=3, c=4)
        self.assertEqual(a[0], 2)
        self.assertEqual(a['a'], 2)
        self.assertEqual(a[1], 3)
        self.assertEqual(a['b'], 3)
        self.assertEqual(list(a.keys()), ['a', 'b', 'c'])
        self.assertEqual(list(a.values()), [2, 3, 4])


class TestFormatting(unittest.TestCase):
    def test_nn3(self):
        fullpath = os.path.join(folder_nn3, 'example.in')
        self.assertTrue(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertFalse(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negfpp_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), 'nextnano3')
        self.assertEqual(defaults.get_fmt('nextnano3')['var_char'], '%')
        self.assertEqual(defaults.get_fmt('nextnano3')['com_char'], '!')
        self.assertEqual(defaults.get_fmt('nextnano3')['input_pattern'], '$end_simulation-dimension')

    def test_nnp(self):
        fullpath = os.path.join(folder_nnp, 'example.in')
        self.assertFalse(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertTrue(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negfpp_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), 'nextnano++')
        self.assertEqual(defaults.get_fmt('nextnano++')['var_char'], '$')
        self.assertEqual(defaults.get_fmt('nextnano++')['com_char'], '#')
        self.assertEqual(defaults.get_fmt('nextnano++')['input_pattern'], 'global{')

    def test_negf(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        self.assertFalse(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertFalse(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertTrue(negf_defaults.is_negf_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negfpp_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), 'nextnano.NEGF') 
        self.assertEqual(defaults.get_fmt('nextnano.NEGF')['var_char'], '$')
        self.assertEqual(defaults.get_fmt('nextnano.NEGF')['com_char'], '<!--')
        self.assertEqual(defaults.get_fmt('nextnano.NEGF')['input_pattern'], '<Simulation')

    # def test_negfpp(self):
    #     # TODO: prepare an NEGF++ .negf input example for test
    #     fullpath = os.path.join(folder_negfpp, 'example.negf')
    #     self.assertFalse(nn3_defaults.is_nn3_input_file(fullpath))
    #     self.assertFalse(nnp_defaults.is_nnp_input_file(fullpath))
    #     self.assertFalse(negf_defaults.is_negf_input_file(fullpath))
    #     self.assertTrue(negf_defaults.is_negfpp_input_file(fullpath))
    #     self.assertEqual(defaults.input_file_type(fullpath), 'nextnano.NEGF++') 
    #     self.assertEqual(defaults.get_fmt('nextnano.NEGF++')['var_char'], '$')
    #     self.assertEqual(defaults.get_fmt('nextnano.NEGF++')['com_char'], '#')
    #     self.assertEqual(defaults.get_fmt('nextnano.NEGF++')['input_pattern'], 'nextnano.NEGF{')


if __name__ == '__main__':
    unittest.main()



import numpy as np
from nextnanopy.utils.mycollections import DictList
from nextnanopy.outputs import Output, AvsAscii, Vtk, DataFileTemplate
from nextnanopy.nnp.defaults import parse_nnp_variable, is_nnp_variable, InputVariable_nnp
from nextnanopy.utils.datasets import Variable, Coord
from nextnanopy.utils.formatting import best_str_to_name_unit


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath, product='nextnano++')
        self.load(**loader_kwargs)

    def get_loader(self):
        if self.extension in ['.v', '.fld', '.coord']:
            loader = AvsAscii
        elif self.extension == '.vtr':
            loader = Vtk
        elif self.extension == '.txt':
            loader = self._find_txt_loader()
        elif self.extension == '.dat':
            loader = Dat
        else:
            raise NotImplementedError(f'Loading datafile with extension {self.extension} is not implemented yet')
        return loader

    def _find_txt_loader(self):
        if self.filename_only in ['variables_input', 'variables_database']:
            loader = InputVariables
        elif self.filename_only == 'materials':
            raise NotImplementedError(f'Loading materials.txt is not implemented yet')
        elif self.filename_only == 'total_charges':
            loader = TotalCharges
        else:
            raise NotImplementedError(f'Datafile {self.filename_only}.txt is not valid')
        return loader


class InputVariables(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.load_raw()
        self.load_variables()

    def load_raw(self):
        with open(self.fullpath, 'r') as f:
            self.raw_lines = f.readlines()

    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            if not is_nnp_variable(line):
                continue
            name, value, comment = parse_nnp_variable(line)
            var = InputVariable_nnp(name=name, value=value, comment=comment, metadata={'line_idx': i})
            variables[var.name] = var
        self.variables = variables
        return self.variables


class TotalCharges(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.load_raw()
        self.load_variables()

    def load_raw(self):
        with open(self.fullpath, 'r') as f:
            self.raw_lines = f.readlines()

    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            try:
                name, value, unit = line.split()
            except Exception as e:
                raise ValueError(f"An error occurred while loading total_charges file:{e}")

            try:
                value = float(value)
            except Exception as e:
                raise ValueError(f"An error occurred while loading total_charges file:{e}")

            # temp solution to omit ":" in the end of the line
            if name[-1] == ':':
                name = name[:-1]

            var = Variable(name=name, unit=unit, value=value)
            variables[var.name] = var

        self.variables = variables
        return self.variables


class Dat(Output):
    def __init__(self, fullpath, FirstVarIsCoordFlag=True):
        super().__init__(fullpath)
        self.load(FirstVarIsCoordFlag)

    def load(self, FirstVarIsCoordFlag):
        self.load_metadata(FirstVarIsCoordFlag)
        self.load_data()

    def _get_headers(self):
        headers = []
        with open(self.fullpath, 'r') as f:
            for line in f:
                try:
                    float(line.split()[0])
                    break
                except:
                    headers.append(line)
        return headers

    def load_metadata(self, FirstVarIsCoordFlag):
        metadata = {}
        headers = self._get_headers()
        metadata['headers'] = headers
        metadata['skip_rows'] = len(headers)

        if len(headers) == 0:
            raise NotImplementedError('.dat file without header')
        else:
            header = headers[-1]  # take the last one by default
        header = header.split()
        ndim = 0
        dkeys = []

        # FirstVarIsCoordFlag = True
        for i, column in enumerate(header):
            key, unit = best_str_to_name_unit(column, default_unit=None)
            metadata[i] = {'name': key, 'unit': unit}
            # if key.lower() in ['x', 'y', 'z', 'position']:
            #    ndim += 1
            #    dkeys.append(i)
            if FirstVarIsCoordFlag:
                ndim += 1
                dkeys.append(i)
                FirstVarIsCoordFlag = False #use this to change behaviour of recognition of coords and variables
        metadata['ndim'] = ndim
        metadata['dkeys'] = dkeys
        self.metadata.update(metadata)
        return metadata

    def load_data(self):
        data = []
        meta = self.metadata
        with open(self.fullpath, 'r') as f:
            for i, line in enumerate(f):
                if i < meta['skip_rows']:
                    continue
                line = line.replace('\n', '').strip().split()
                if line:
                    data.append(line)
        data = np.array(data, dtype=float).T  # columns 1st index
        coords, variables = DictList(), DictList()
        dims = []
        for i, values in enumerate(data):
            vm = meta[i]
            if i in meta['dkeys']:
                #values = np.unique(values)
                dims.append(values.size)
                var = Coord(name=vm['name'], unit=vm['unit'], dim=i, value=values)
                coords[var.name] = var
            else:
                if dims:
                    values = values.reshape(*dims)
                var = Variable(name=vm['name'], unit=vm['unit'], value=values)
                variables[var.name] = var
        self.coords = coords
        self.variables = variables
        return coords, variables


    def save(self, new_location, extension='dat'):
        with open(new_location, 'w') as f:
            # Write headers
            for header in self.metadata['headers']:
                f.write(header)

            # Write data
            combined_data = [coord.value for coord in self.coords] + [variable.value for variable in self.variables]

            data = np.column_stack(combined_data).transpose()
            np.savetxt(f, data.T)

        # Optionally update the fullpath attribute to the new location
        self.fullpath = new_location



import os
from pathlib import Path
from nextnanopy.utils.config import Config

products = ['nextnano++', 'nextnano3', 'nextnano.NEGF', 'nextnano.NEGF++', 'nextnano.MSB', 'nextnanoevo']
default_folder = str(Path.home())
config_default_path = os.path.join(default_folder, '.nextnanopy-config')
messages = {
    'load_input': [None, None],
    'save_input': [None, None],
    'execute_input': [None, None],
    'load_output': [None, None],
}


def get_InputFile(product):
    if product == 'nextnano++':
        from nextnanopy.nnp.inputs import InputFile
    elif product == 'nextnano3':
        from nextnanopy.nn3.inputs import InputFile
    elif product == 'nextnano.NEGF':
        from nextnanopy.negf.inputs import InputFile
    elif product == 'nextnano.NEGF++':
        from nextnanopy.negf.inputspp import InputFile
    elif product == 'nextnano.MSB':
        from nextnanopy.msb.inputs import InputFile
    elif product == 'nextnanoevo':
        raise ValueError('There is no InputFile format for nextnanoevo')
    elif product == 'not valid':
        from nextnanopy.inputs import InputFileTemplate as InputFile
    else:
        raise ValueError(f'{product} is not valid')
    return InputFile


def get_DataFile(product):
    if product == 'nextnano3':
        from nextnanopy.nn3.outputs import DataFile
    elif product == 'nextnano++':
        from nextnanopy.nnp.outputs import DataFile
    elif product == 'nextnano.NEGF' or 'nextnano.NEGF++':
        from nextnanopy.negf.outputs import DataFile
    elif product == 'nextnano.MSB':
        from nextnanopy.negf.outputs import DataFile
    elif product == 'nextnanoevo':
        raise ValueError('There is no DataFile format for nextnanoevo')
    else:
        raise ValueError(f'{product} is not valid')
    return DataFile


def get_command(product):
    if product == 'nextnano++':
        from nextnanopy.nnp.defaults import command_nnp as command
    elif product == 'nextnano3':
        from nextnanopy.nn3.defaults import command_nn3 as command
    elif product == 'nextnano.NEGF':
        from nextnanopy.negf.defaults import command_negf as command
    elif product == 'nextnano.NEGF++':
        from nextnanopy.negf.defaults import command_negfpp as command
    elif product == 'nextnano.MSB':
        from nextnanopy.msb.defaults import command_msb as command
    elif product == 'nextnanoevo':
        raise ValueError('There is no command format for nextnanoevo')
    else:
        raise ValueError(f'{product} is not valid')
    return command


def get_fmt(product):
    if product == 'nextnano++':
        from nextnanopy.nnp.defaults import fmt
    elif product == 'nextnano3':
        from nextnanopy.nn3.defaults import fmt
    elif product == 'nextnano.NEGF':
        from nextnanopy.negf.defaults import fmt
    elif product == 'nextnano.NEGF++':
        from nextnanopy.negf.defaults import fmtpp
    elif product == 'nextnano.MSB':
        from nextnanopy.msb.defaults import fmt
    elif product == 'nextnanoevo':
        raise ValueError('There is no formatting defaults for nextnanoevo')
    else:
        raise ValueError(f'{product} is not valid')
    return fmt


def input_file_type(fullpath):
    from nextnanopy.nnp.defaults import is_nnp_input_file
    from nextnanopy.nn3.defaults import is_nn3_input_file
    from nextnanopy.negf.defaults import is_negf_input_file, is_negfpp_input_file
    from nextnanopy.msb.defaults import is_msb_input_file
    if is_nn3_input_file(fullpath):
        return 'nextnano3'
    elif is_nnp_input_file(fullpath):
        return 'nextnano++'
    elif is_negf_input_file(fullpath):
        return 'nextnano.NEGF'
    elif is_negfpp_input_file(fullpath):
        return 'nextnano.NEGF++'
    elif is_msb_input_file(fullpath):
        return 'nextnano.MSB'
    else:
        return 'not valid'


def input_text_type(text):
    from nextnanopy.nnp.defaults import is_nnp_input_text
    from nextnanopy.nn3.defaults import is_nn3_input_text
    from nextnanopy.negf.defaults import is_negf_input_text, is_negfpp_input_text
    from nextnanopy.msb.defaults import is_msb_input_text
    if is_nn3_input_text(text):
        return 'nextnano3'
    elif is_nnp_input_text(text):
        return 'nextnano++'
    elif is_negf_input_text(text):
        return 'nextnano.NEGF'
    elif is_negfpp_input_text(text):
        return 'nextnano.NEGF++'
    elif is_msb_input_text(text):
        return 'nextnano.MSB'
    else:
        return 'not valid'


def get_config_validators():
    config_validator = {product: _get_config_validator(product) for product in products}
    return config_validator


def get_config_defaults():
    config_defaults = {product: _get_config_default(product) for product in products}
    return config_defaults


def _get_config_validator(product):
    if product == 'nextnano++':
        from nextnanopy.nnp.defaults import config_validator
    elif product == 'nextnano3':
        from nextnanopy.nn3.defaults import config_validator
    elif product == 'nextnano.NEGF' or product == 'nextnano.NEGF++':
        from nextnanopy.negf.defaults import config_validator
    elif product == 'nextnano.MSB':
        from nextnanopy.msb.defaults import config_validator
    elif product == 'nextnanoevo':
        from nextnanopy.nnevo.defaults import config_validator
    else:
        raise ValueError(f'{product} is not valid')
    return config_validator


def _get_config_default(product):
    if product == 'nextnano++':
        from nextnanopy.nnp.defaults import config_default
    elif product == 'nextnano3':
        from nextnanopy.nn3.defaults import config_default
    elif product == 'nextnano.NEGF' or product == 'nextnano.NEGF++':
        from nextnanopy.negf.defaults import config_default
    elif product == 'nextnano.MSB':
        from nextnanopy.msb.defaults import config_default
    elif product == 'nextnanoevo':
        from nextnanopy.nnevo.defaults import config_default
    else:
        raise ValueError(f'{product} is not valid')
    return config_default


class NNConfig(Config):
    def __init__(self, fullpath=None):
        self.default_fullpath = config_default_path
        validators = get_config_validators()
        self.defaults = get_config_defaults()
        if fullpath is None:
            fullpath = self.default_fullpath
        super().__init__(fullpath, validators)
        if not os.path.isfile(fullpath):
            self.to_default()
            self.save()
        elif not self._complete:
            # TODO add tests for these behaviour
            # ensures smooth transition to additional product: NEGF++
            self.update_with_defaults()
            self.save()


    def to_default(self):
        for section in self.defaults.keys():
            if section not in self.sections:
                self.add_section(section)
            for option, value in self.defaults[section].items():
                self.set(section, option, value)

    def reset(self):
        self.to_default()
        self.save()


    @property
    def _complete(self):
        return self.check_complete()

    def check_complete(self):
        for section in self.defaults.keys():
            if section not in self.sections:
                return False
        return True

    def update_with_defaults(self):
        for section in self.defaults.keys():
            if section not in self.sections:
                self.add_section(section)
                for option, value in self.defaults[section].items():
                    self.set(section, option, value)



from nextnanopy.nnp.assistants import InputAssistant
from collections import OrderedDict
from nextnanopy.shapes import GdsPolygonsRaw
import numpy as np
import warnings

class GdsPolygons(GdsPolygonsRaw):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_obelisks(self, zi, zf):
        warnings.warn("The usage of get_obelisks is deprecated. Please, GdsPolygons.get_polygonal_prisms instead", DeprecationWarning)
        shapes = []
        z = np.array([zi, zf] * 2)
        for si in self.slices:
            xs, ys = si.correct_xy()
            for x, y in zip(xs, ys):
                kwargs = {
                    'base_x': x[:2],
                    'base_y': y[:2],
                    'base_z': z[:2],
                    'top_x': x[2:],
                    'top_y': y[2:],
                    'top_z': z[2:],
                }
                shapes.append(Obelisk(**kwargs))
        return shapes

    def get_polygonal_prisms(self, zi, zf):
        z = np.array([zi, zf])
        axes = ['x', 'y', 'z']
        shapes = [Polygonal_prism(axes=axes, vertexes=xy, height=z) for xy in
                  self.polygons_xy]
        return shapes


class Shape(object):
    def __init__(self):
        self._ia = InputAssistant()

    @property
    def text(self):
        return ''

    @property
    def preview(self):
        self._ia.preview(self.text, nums=False)


class Obelisk(Shape):

    def __init__(self, base_x, base_y, base_z, top_x, top_y, top_z):
        super().__init__()
        self.base_x = base_x
        self.base_y = base_y
        self.base_z = base_z
        self.top_x = top_x
        self.top_y = top_y
        self.top_z = top_z

    @property
    def text(self):
        return self._ia.region_obelisk(**self.kwargs)

    @property
    def kwargs(self):
        _kwargs = OrderedDict()
        keys = ['base_x', 'base_y', 'base_z', 'top_x', 'top_y', 'top_z']
        for key in keys:
            if key not in dir(self):
                raise KeyError(f'{key} is not defined in the attributes')
            else:
                _kwargs[key] = getattr(self, key)
        return _kwargs


class Polygonal_prism(Shape):

    def __init__(self, axes=['x', 'y', 'z'], vertexes=[[10.5, 14.0]],
                 height=[0, 10]):
        super().__init__()
        self.axes = axes
        self.vertexes = vertexes
        self.height = height

    @property
    def text(self):
        return self._ia.region_polygonal_prism(**self.kwargs)

    @property
    def kwargs(self):
        _kwargs = OrderedDict()
        keys = ['axes', 'vertexes', 'height']
        for key in keys:
            if key not in dir(self):
                raise KeyError(f'{key} is not defined in the attributes')
            else:
                _kwargs[key] = getattr(self, key)
        return _kwargs



import unittest
from nextnanopy.shapes import GdsPolygonsRaw, units_factor, validate_unit
import numpy as np
import os

np_equal = np.testing.assert_almost_equal

folder = os.path.join('tests', 'gds')


class TestShapes(unittest.TestCase):
    def test_unit(self):
        self.assertEqual(units_factor['nm'], 1e-9)
        self.assertEqual(units_factor['um'], 1e-6)
        self.assertEqual(units_factor['mm'], 1e-3)
        self.assertEqual(units_factor['m'], 1)
        self.assertEqual(units_factor['si'], 1)
        self.assertRaises(KeyError, validate_unit, 'none')
        self.assertTrue(validate_unit('NM'))

    def test_example0(self):
        fullpath = os.path.join(folder, 'example0.gds')
        gpols = GdsPolygonsRaw(fullpath, unit='nm')
        self.assertEqual(gpols.fullpath, fullpath)
        np_equal(gpols.labels, [0])
        self.assertEqual(gpols.unit, 'nm')
        self.assertEqual(gpols.nb_polygons, 1)
        np_equal(gpols.xy[0][0], np.array([-500, -500, 500, 500]))
        np_equal(gpols.xy[0][1], np.array([-500, 500, 500, -500]))

        gpols.unit = 'm'
        self.assertEqual(gpols.unit, 'm')
        self.assertEqual(gpols.nb_polygons, 1)
        np_equal(gpols.xy[0][0], np.array([-500, -500, 500, 500]) * 1e-9)
        np_equal(gpols.xy[0][1], np.array([-500, 500, 500, -500]) * 1e-9)

        gpols.labels = [2]
        np_equal(gpols.labels, [2])

        gpols.labels = ['2']
        self.assertEqual(gpols.labels, ['2'])

    def test_example1(self):
        fullpath = os.path.join(folder, 'example1.gds')
        gpols = GdsPolygonsRaw(fullpath, unit='nm')
        np_equal(gpols.labels, [0, 1, 2, 3])
        self.assertEqual(gpols.nb_polygons, 4)
        np_equal(gpols.xy[0][0], np.array([1500., 1500., 1975., 1975., 2025., 2025., 2500., 2500.]))

    def test_example2(self):
        fullpath = os.path.join(folder, 'example2.gds')
        gpols = GdsPolygonsRaw(fullpath, unit='nm')
        np_equal(gpols.labels, [0, 1])
        self.assertEqual(gpols.nb_polygons, 2)


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from nextnanopy.inputs import InputFile, Sweep


def delete_files(start,directory = os.getcwd(),exceptions =None):
    for fname in os.listdir(directory):
        if fname.startswith(start):
            if exceptions:
                if fname in exceptions:
                    continue
                else:
                    os.remove(os.path.join(directory, fname))
            else:
                os.remove(os.path.join(directory, fname))


folder_nnp = os.path.join('tests', 'datafiles', 'nextnano++')
folder_nn3 = os.path.join('tests', 'datafiles', 'nextnano3')
folder_negf = os.path.join('tests', 'datafiles', 'nextnano.NEGF')
folder_negfpp = os.path.join('tests', 'datafiles', 'nextnano.NEGF++')


class Test_nnp(unittest.TestCase):
    def test_load(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')

        file = InputFile(fullpath)
        self.assertEqual(file.product, 'nextnano++')

        self.assertEqual(len(file.variables.keys()), 7)
        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, float(0.0))
        self.assertEqual(file.variables['float'].comment, 'float number (a.u.)')

        self.assertEqual(file.variables['int'].name, 'int')
        self.assertEqual(file.variables['int'].value, int(0))
        self.assertEqual(file.variables['int'].comment, 'int number (random = 3)')

        self.assertEqual(file.variables['str'].name, 'str')
        self.assertEqual(file.variables['str'].value, '"test"')
        self.assertEqual(file.variables['str'].comment, 'first comments # second comment')

        self.assertEqual(file.variables['reference'].name, 'reference')
        self.assertEqual(file.variables['reference'].value, '$int - 1')

        self.assertEqual(file.variables['NoSpacing'].name, 'NoSpacing')
        self.assertEqual(file.variables['NoSpacing'].value, int('7'))
        self.assertEqual(file.variables['NoSpacing'].comment, 'this is a comment')
        self.assertEqual(file.variables['NoSpacing'].text, '$NoSpacing = 7 # this is a comment')

        self.assertEqual(file.variables['MAYUS'].name, 'MAYUS')
        self.assertEqual(file.variables['MAYUS'].value, 'TEXT')
        self.assertEqual(file.variables['MAYUS'].comment, '')
        self.assertEqual(file.variables['MAYUS'].text, '$MAYUS = TEXT')

        fullpath = os.path.join(folder_nnp, 'virtual_file.in')
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)

        self.assertEqual(file.variables['float'], file.get_variable('float'))
        self.assertRaises(KeyError, file.get_variable, name='new_variable')

    def test_set_variables(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)
        file.set_variable('float', 1e-5, 'some comment', 'some unit')#unit is not visible in nn++

        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, 1e-5)
        self.assertEqual(file.variables['float'].comment, 'some comment')
        self.assertEqual(file.variables['float'].text, f'$float = {str(1e-5)} # some comment')
        self.assertEqual(file.lines[3], f'$float = {str(1e-5)} # some comment')

        file.set_variable('float', value=1e-7)
        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, 1e-7)
        self.assertEqual(file.variables['float'].comment, 'some comment')
        self.assertEqual(file.variables['float'].text, f'$float = {str(1e-7)} # some comment')
        self.assertEqual(file.lines[3], f'$float = {str(1e-7)} # some comment')

        file.set_variable('float', comment='new comment')
        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, 1e-7)
        self.assertEqual(file.variables['float'].comment, 'new comment')
        self.assertEqual(file.variables['float'].text, f'$float = {str(1e-7)} # new comment')
        self.assertEqual(file.lines[3], f'$float = {str(1e-7)} # new comment')

        file.set_variable('float', value=0)
        self.assertEqual(file.variables['float'].value, 0)
        file.set_variable('float', value='0')
        self.assertEqual(file.variables['float'].value, '0')
        self.assertRaises(KeyError, file.set_variable, name='new_variable')
        self.assertRaises(KeyError, file.set_variable, name='new_variable')

    def test_fullpath(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)

        self.assertEqual(file.fullpath, fullpath)
        self.assertEqual(file.save(file.fullpath, overwrite=False),
                         os.path.join(folder_nnp, 'only_variables_0.in'))
        os.remove(file.fullpath)

    def test_config(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)
        from nextnanopy import config
        for key, value in config.config['nextnano++'].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_text(self):
        fullpath_onlyvar = os.path.join(folder_nnp, 'only_variables.in')
        fullpath_example = os.path.join(folder_nnp, 'example.in')
        file = InputFile(fullpath_onlyvar)
        text = file.text

        new_file = InputFile(fullpath_example)
        # new_file = InputFile()
        # self.assertEqual(new_file.product, 'not valid')
        # self.assertEqual(new_file.fullpath, None)
        # self.assertEqual(new_file.text, '')
        # self.assertEqual(new_file.raw_lines, [])

        new_file.text = text
        self.assertEqual(new_file.product, 'nextnano++')
        # self.assertEqual(new_file.fullpath, None)
        self.assertEqual(new_file.text, text)
        self.assertEqual(new_file.lines, file.lines)
        self.assertEqual(new_file.variables['MAYUS'].name, 'MAYUS')
        self.assertEqual(new_file.variables['MAYUS'].value, 'TEXT')
        self.assertEqual(new_file.variables['MAYUS'].comment, '')
        self.assertEqual(new_file.variables['MAYUS'].text, '$MAYUS = TEXT')
        # self.assertRaises(ValueError, new_file.save)
        self.assertEqual(new_file.save(file.fullpath, overwrite=False),
                         os.path.join(folder_nnp, 'only_variables_0.in'))
        os.remove(new_file.fullpath)

    def test_set_and_save(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)
        file.set_variable(name =  'float', value = 0.4)
        self.assertAlmostEqual(file.variables['float'].value, 0.4)

        self.addCleanup(os.remove,os.path.join(folder_nnp,'only_variables_0.in'))
        file.save()
        self.assertTrue(os.path.isfile(os.path.join(folder_nnp,'only_variables_0.in')))


    def test_same_dir_saving(self):
        current_directory = os.getcwd()
        self.addCleanup(os.chdir,current_directory)

        nnp_datafile_dir = os.path.join(folder_nnp)
        os.chdir(nnp_datafile_dir)

        path = 'only_variables.in'
        file = InputFile(path)
        file.set_variable(name='float', value=0.3333)
        self.assertAlmostEqual(file.variables['float'].value, 0.3333)
        self.addCleanup(delete_files,'only_variables', directory='.', exceptions=['only_variables.in'])
        file.save()
        self.assertTrue(os.path.isfile('only_variables_0.in'))


    ###content tests

    def test_content_get(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)


        self.assertIsNotNone(file.content)
        self.assertEqual(file.content[0],'$float = 0.0 ')
        self.assertEqual(file.content[-1].name, 'global')

    def test_content_set(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)


        file.content[0] = '$DUMMY = 1'
        self.assertEqual(file.content[0],'$DUMMY = 1')
        file.content['_entry_0'] = 'DUMMY LINE'
        self.assertEqual(file.content[0], 'DUMMY LINE')




class Test_nn3(unittest.TestCase):
    def test_load(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')

        file = InputFile(fullpath)
        self.assertEqual(file.product, 'nextnano3')

        self.assertEqual(len(file.variables.keys()), 7)
        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, float(0.0))
        self.assertEqual(file.variables['float'].comment, 'float number (a.u.)')

        self.assertEqual(file.variables['int'].name, 'int')
        self.assertEqual(file.variables['int'].value, int(0))
        self.assertEqual(file.variables['int'].comment, 'int number (random = 3)')

        self.assertEqual(file.variables['str'].name, 'str')
        self.assertEqual(file.variables['str'].value, '"test"')
        self.assertEqual(file.variables['str'].comment, 'first comments # second comment')

        self.assertEqual(file.variables['reference'].name, 'reference')
        self.assertEqual(file.variables['reference'].value, '$int - 1')

        self.assertEqual(file.variables['NoSpacing'].name, 'NoSpacing')
        self.assertEqual(file.variables['NoSpacing'].value, int('7'))
        self.assertEqual(file.variables['NoSpacing'].comment, 'this is a comment')

        self.assertEqual(file.variables['MAYUS'].name, 'MAYUS')
        self.assertEqual(file.variables['MAYUS'].value, 'TEXT')
        self.assertEqual(file.variables['MAYUS'].comment, '')

        fullpath = os.path.join(folder_nn3, 'virtual_file.in')
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)

        self.assertEqual(file.variables['float'], file.get_variable('float'))
        self.assertRaises(KeyError, file.get_variable, name='new_variable')

    def test_set_variables(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)
        file.set_variable('float', 1e-5, 'some comment')

        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, 1e-5)
        self.assertEqual(file.variables['float'].comment, 'some comment')
        self.assertEqual(file.variables['float'].text, f'%float = {str(1e-5)} ! some comment')
        self.assertEqual(file.lines[3], f'%float = {str(1e-5)} ! some comment')

        file.set_variable('float', value=1e-7)
        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, 1e-7)
        self.assertEqual(file.variables['float'].comment, 'some comment')
        self.assertEqual(file.variables['float'].text, f'%float = {str(1e-7)} ! some comment')
        self.assertEqual(file.lines[3], f'%float = {str(1e-7)} ! some comment')

        file.set_variable('float', comment='new comment')
        self.assertEqual(file.variables['float'].name, 'float')
        self.assertEqual(file.variables['float'].value, 1e-7)
        self.assertEqual(file.variables['float'].comment, 'new comment')
        self.assertEqual(file.variables['float'].text, f'%float = {str(1e-7)} ! new comment')
        self.assertEqual(file.lines[3], f'%float = {str(1e-7)} ! new comment')

        file.set_variable('float', value=0)
        self.assertEqual(file.variables['float'].value, 0)
        file.set_variable('float', value='0')
        self.assertEqual(file.variables['float'].value, '0')
        self.assertRaises(KeyError, file.set_variable, name='new_variable')

    def test_fullpath(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)

        self.assertEqual(file.fullpath, fullpath)
        self.assertEqual(file.save(file.fullpath, overwrite=False),
                         os.path.join(folder_nn3, 'only_variables_0.in'))
        os.remove(file.fullpath)

    def test_config(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)
        from nextnanopy import config
        for key, value in config.config['nextnano3'].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_save(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)
        new_folder = os.path.join(folder_nn3, 'temp')
        new_file = os.path.join(new_folder, 'example_copy.in')
        self.assertRaises(FileNotFoundError, file.save, new_file, overwrite=True, automkdir=False)
        self.assertEqual(file.save(new_file, overwrite=True, automkdir=True), new_file)
        os.remove(file.fullpath)
        os.rmdir(new_folder)

    def test_text(self):
        fullpath_onlyvar = os.path.join(folder_nn3, 'only_variables.in')
        fullpath_example = os.path.join(folder_nn3, 'example.in')
        file = InputFile(fullpath_onlyvar)
        text = file.text

        new_file = InputFile(fullpath_example)
        # self.assertEqual(new_file.product, 'not valid')
        # self.assertEqual(new_file.fullpath, None)
        # self.assertEqual(new_file.text, '')
        # self.assertEqual(new_file.raw_lines, [])

        new_file.text = text

        self.assertEqual(new_file.product, 'nextnano3')
        # self.assertEqual(new_file.fullpath, None)
        self.assertEqual(new_file.text, text)
        self.assertEqual(new_file.lines, file.lines)
        self.assertEqual(new_file.variables['MAYUS'].name, 'MAYUS')
        self.assertEqual(new_file.variables['MAYUS'].value, 'TEXT')
        self.assertEqual(new_file.variables['MAYUS'].comment, '')
        self.assertEqual(new_file.variables['MAYUS'].text, '%MAYUS = TEXT')
        # self.assertRaises(ValueError, new_file.save)
        self.assertEqual(new_file.save(file.fullpath, overwrite=False),
                         os.path.join(folder_nn3, 'only_variables_0.in'))
        os.remove(new_file.fullpath)

    def test_set_and_save(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)
        file.set_variable(name =  'float', value = 0.4)
        self.assertAlmostEqual(file.variables['float'].value, 0.4)

        self.addCleanup(os.remove,os.path.join(folder_nn3,'only_variables_0.in'))
        file.save()
        self.assertTrue(os.path.isfile(os.path.join(folder_nn3,'only_variables_0.in')))

    def test_same_dir(self):
        current_directory = os.getcwd()
        self.addCleanup(os.chdir,current_directory)

        nn3_datafile_dir = os.path.join(folder_nn3)
        os.chdir(nn3_datafile_dir)

        path = 'only_variables.in'
        file = InputFile(path)
        file.set_variable(name='float', value=0.3333)
        self.assertAlmostEqual(file.variables['float'].value, 0.3333)
        self.addCleanup(delete_files,'only_variables', directory='.', exceptions=['only_variables.in'])
        file.save()
        self.assertTrue(os.path.isfile('only_variables_0.in'))


    def test_content(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        file = InputFile(fullpath)


        self.assertIsNone(file.content)


class Test_negf(unittest.TestCase):
    def test_load(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)

        self.assertEqual(file.product, 'nextnano.NEGF')

        self.assertEqual(len(file.variables.keys()), 4)

        self.assertEqual(file.variables['variable1'].name, 'variable1')
        self.assertEqual(file.variables['variable1'].value, float(0.24))
        self.assertEqual(file.variables['variable1'].comment, 'Some comment')
        self.assertEqual(file.variables['variable1'].unit, 'a.u.')

        self.assertEqual(file.variables['variable2'].name, 'variable2')
        self.assertEqual(file.variables['variable2'].value, int(0))
        self.assertEqual(file.variables['variable2'].comment, 'Another comment')
        self.assertEqual(file.variables['variable2'].unit, 'meV')

        self.assertEqual(file.variables['text_var'].name, 'text_var')
        self.assertEqual(file.variables['text_var'].value, 'some text')

        self.assertEqual(file.variables['ref_var'].value, '$(1-variable1)')




        fullpath = os.path.join(folder_negf, 'virtual_file.xml')
        self.assertRaises(FileNotFoundError, InputFile, fullpath)
    def test_get_variables(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)

        self.assertEqual(file.variables['variable1'], file.get_variable('variable1'))

        self.assertRaises(KeyError, file.get_variable, name='new_variable')

    def test_set_variables(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)

        file.set_variable('variable1', 0.137, 'test comment', 'test unit^2')
        self.assertEqual(file.variables['variable1'].value, float(0.137))
        self.assertEqual(file.variables['variable1'].comment, 'test comment')
        self.assertEqual(file.variables['variable1'].unit, 'test unit^2')
        #self.assertEqual(file.lines[19], '<Name Comment="Some comment">$variable1</Name>')

        file.set_variable('text_var', 'string variable test')
        self.assertEqual(file.variables['text_var'].value, 'string variable test')

        file.set_variable('ref_var', '$variable1 - 0.1')
        self.assertEqual(file.variables['ref_var'].value, '$variable1 - 0.1')

        self.assertRaises(KeyError, file.set_variable, name='new_variable')

    def test_fullpath(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)

        self.assertEqual(file.fullpath, fullpath)
        self.assertEqual(file.save(file.fullpath, overwrite=False),
                         os.path.join(folder_negf, 'example_0.xml'))
        os.remove(file.fullpath)

    def test_config(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)
        from nextnanopy import config
        for key, value in config.config['nextnano.NEGF'].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_save(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)

        new_folder = os.path.join(folder_negf, 'temp')
        new_file = os.path.join(new_folder, 'example_copy.in')
        self.assertRaises(FileNotFoundError, file.save, new_file, overwrite=True, automkdir=False)
        self.assertEqual(file.save(new_file, overwrite=True, automkdir=True), new_file)
        os.remove(file.fullpath)
        os.rmdir(new_folder)

    def test_set_and_save(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)
        file.set_variable(name =  'variable1', value = 0.4)
        self.assertAlmostEqual(file.variables['variable1'].value, 0.4)

        self.addCleanup(os.remove,os.path.join(folder_negf,'example_0.xml'))
        file.save()
        self.assertTrue(os.path.isfile(os.path.join(folder_negf,'example_0.xml')))


    def test_content(self):
        fullpath = os.path.join(folder_negf, 'example.xml')
        file = InputFile(fullpath)


        self.assertIsNone(file.content)


class Test_negfpp(unittest.TestCase):
    # TODO: implement test for NEGF++ input file
    def test_load(self):
        fullpath = os.path.join(folder_negf, 'Minimal_InputFile.negf')

        file = InputFile(fullpath)
        self.assertEqual(file.product, 'nextnano.NEGF++')

        self.assertEqual(len(file.variables.keys()), 1)
        self.assertEqual(file.variables['alloyComposition'].name, 'alloyComposition')
        self.assertAlmostEqual(file.variables['alloyComposition'].value, float(0.15), delta=1e-9)
        self.assertEqual(file.variables['alloyComposition'].comment, 'alloy composition')



        fullpath = os.path.join(folder_nnp, 'virtual_file.in')
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = os.path.join(folder_negf, 'Minimal_InputFile.negf')
        file = InputFile(fullpath)

        self.assertEqual(file.variables['alloyComposition'], file.get_variable('alloyComposition'))
        self.assertRaises(KeyError, file.get_variable, name='new_variable')

    def test_set_variables(self):
        fullpath = os.path.join(folder_negf, 'Minimal_InputFile.negf')
        file = InputFile(fullpath)
        file.set_variable('alloyComposition', 1e-5, 'some comment', 'some unit')#unit is not visible in nn++

        self.assertEqual(file.variables['alloyComposition'].name, 'alloyComposition')
        self.assertEqual(file.variables['alloyComposition'].value, 1e-5)
        self.assertEqual(file.variables['alloyComposition'].comment, 'some comment')
        self.assertEqual(file.variables['alloyComposition'].text, f'$alloyComposition = {str(1e-5)} # some comment')
        self.assertEqual(file.lines[0], f'$alloyComposition = {str(1e-5)} # some comment')

        file.set_variable('alloyComposition', value=1e-7)
        self.assertEqual(file.variables['alloyComposition'].name, 'alloyComposition')
        self.assertEqual(file.variables['alloyComposition'].value, 1e-7)
        self.assertEqual(file.variables['alloyComposition'].comment, 'some comment')
        self.assertEqual(file.variables['alloyComposition'].text, f'$alloyComposition = {str(1e-7)} # some comment')
        self.assertEqual(file.lines[0], f'$alloyComposition = {str(1e-7)} # some comment')

        file.set_variable('alloyComposition', comment='new comment')
        self.assertEqual(file.variables['alloyComposition'].name, 'alloyComposition')
        self.assertEqual(file.variables['alloyComposition'].value, 1e-7)
        self.assertEqual(file.variables['alloyComposition'].comment, 'new comment')
        self.assertEqual(file.variables['alloyComposition'].text, f'$alloyComposition = {str(1e-7)} # new comment')
        self.assertEqual(file.lines[0], f'$alloyComposition = {str(1e-7)} # new comment')

        file.set_variable('alloyComposition', value=0)
        self.assertEqual(file.variables['alloyComposition'].value, 0)

        self.assertRaises(KeyError, file.set_variable, name='new_variable')
        self.assertRaises(KeyError, file.set_variable, name='new_variable')

    def test_fullpath(self):
        fullpath = os.path.join(folder_negf, 'Minimal_InputFile.negf')
        file = InputFile(fullpath)

        self.assertEqual(file.fullpath, fullpath)
        self.assertEqual(file.save(file.fullpath, overwrite=False),
                         os.path.join(folder_negf, 'Minimal_InputFile_0.negf'))
        os.remove(file.fullpath)

    def test_config(self):
        fullpath = os.path.join(folder_negf, 'Minimal_InputFile.negf')
        file = InputFile(fullpath)
        from nextnanopy import config
        for key, value in config.config['nextnano.NEGF++'].items():
            self.assertEqual(file.default_command_args[key], value)

    # def test_text(self):
    #     fullpath = os.path.join(folder_nnp, 'only_variables.in')
    #
    #     file = InputFile(fullpath)
    #     text = file.text
    #
    #     new_file = InputFile()
    #     self.assertEqual(new_file.product, 'not valid')
    #     self.assertEqual(new_file.fullpath, None)
    #     self.assertEqual(new_file.text, '')
    #     self.assertEqual(new_file.raw_lines, [])
    #
    #     new_file.text = text
    #     self.assertEqual(new_file.product, 'nextnano++')
    #     self.assertEqual(new_file.fullpath, None)
    #     self.assertEqual(new_file.text, text)
    #     self.assertEqual(new_file.lines, file.lines)
    #     self.assertEqual(new_file.variables['MAYUS'].name, 'MAYUS')
    #     self.assertEqual(new_file.variables['MAYUS'].value, 'TEXT')
    #     self.assertEqual(new_file.variables['MAYUS'].comment, '')
    #     self.assertEqual(new_file.variables['MAYUS'].text, '$MAYUS = TEXT')
    #     self.assertRaises(ValueError, new_file.save)
    #     self.assertEqual(new_file.save(file.fullpath, overwrite=False),
    #                      os.path.join(folder_nnp, 'only_variables_0.in'))
    #     os.remove(new_file.fullpath)

    def test_set_and_save(self):
        fullpath = os.path.join(folder_negf, 'Minimal_InputFile.negf')
        file = InputFile(fullpath)
        file.set_variable(name='alloyComposition', value=0.4)
        self.assertAlmostEqual(file.variables['alloyComposition'].value, 0.4)

        self.addCleanup(os.remove,os.path.join(folder_negf,'Minimal_InputFile_0.negf'))
        file.save()
        self.assertTrue(os.path.isfile(os.path.join(folder_negf,'Minimal_InputFile_0.negf')))


    def test_same_dir_saving(self):
        current_directory = os.getcwd()
        self.addCleanup(os.chdir,current_directory)


        os.chdir(folder_negf)

        path = 'Minimal_InputFile.negf'
        file = InputFile(path)
        file.set_variable(name='alloyComposition', value=0.3333)
        self.assertAlmostEqual(file.variables['alloyComposition'].value, 0.3333)
        self.addCleanup(delete_files,'Minimal_InputFile', directory='.', exceptions=['Minimal_InputFile.negf'])
        file.save()
        self.assertTrue(os.path.isfile('Minimal_InputFile_0.negf'))


    ###content tests

    # def test_content_get(self): # LATER also check this
    #     fullpath = os.path.join(folder_nnp, 'only_variables.in')
    #     file = InputFile(fullpath)
    #
    #
    #     self.assertIsNotNone(file.content)
    #     self.assertEqual(file.content[0],'$float = 0.0 ')
    #     self.assertEqual(file.content[-1].name, 'global')
    #
    # def test_content_set(self):
    #     fullpath = os.path.join(folder_nnp, 'only_variables.in')
    #     file = InputFile(fullpath)
    #
    #
    #     file.content[0] = '$DUMMY = 1'
    #     self.assertEqual(file.content[0],'$DUMMY = 1')
    #     file.content['_entry_0'] = 'DUMMY LINE'
    #     self.assertEqual(file.content[0], 'DUMMY LINE')

class TestInputFile(unittest.TestCase):

    def test_access_by_index(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)
        for key, value in file.variables.items():
            self.assertEqual(file[key], value)

    def test_for_loop(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)
        for i, fi in enumerate(file):
            self.assertEqual(file.variables[i], fi)

    def test_fullpath(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        file = InputFile(fullpath)
        self.assertEqual(file.fullpath, fullpath)
        self.assertEqual(file.filename, 'only_variables.in')
        self.assertEqual(file.filename_only, 'only_variables')
        self.assertEqual(file.folder_input, folder_nnp)
        self.assertEqual(file.execute_info, {})

        file.filename = 'new_name.in'
        self.assertEqual(file.filename, 'new_name.in')
        self.assertEqual(file.filename_only, 'new_name')
        self.assertEqual(file.fullpath, os.path.join(folder_nnp, 'new_name.in'))
        self.assertEqual(file.folder_input, folder_nnp)

        file.filename_only = 'another_filename'
        self.assertEqual(file.filename, 'another_filename.in')
        self.assertEqual(file.filename_only, 'another_filename')
        self.assertEqual(file.fullpath, os.path.join(folder_nnp, 'another_filename.in'))
        self.assertEqual(file.folder_input, folder_nnp)

        npath = os.path.join('new','folder')
        file.folder_input = npath
        self.assertEqual(file.filename, 'another_filename.in')
        self.assertEqual(file.filename_only, 'another_filename')
        self.assertEqual(file.fullpath, os.path.join(npath, 'another_filename.in'))
        self.assertEqual(file.folder_input, npath)

        npath = os.path.join('random', 'path')
        file.execute_info['outputdirectory'] = npath
        self.assertEqual(file.folder_output, npath)

class TestSweep(unittest.TestCase):
    def test_init(self):
        self.assertRaises(TypeError, Sweep)
        sweep = Sweep({})
        self.assertEqual(sweep.fullpath, None)
        self.assertEqual(sweep.var_sweep, {})
        self.assertRaises(ValueError, Sweep, {'Name': 'some_name'})



    def test_nnp_init(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        self.assertRaises(ValueError, Sweep, {'Name': 'some_name'}, fullpath)

        sweep = Sweep({},fullpath)
        self.assertEqual(sweep.fullpath, fullpath)
        self.assertEqual(sweep.var_sweep, {})
        self.assertFalse(sweep.input_files)
        self.assertFalse(sweep.sweep_output_directory)


        self.assertRaises(ValueError, Sweep, {'float':1})
        self.assertRaises(TypeError, Sweep, {'float': 1}, fullpath)

        sweep = Sweep({'float': [1,2,5]}, fullpath)
        self.assertEqual(sweep.fullpath, fullpath)
        self.assertEqual(sweep.var_sweep['float'],[1,2,5])

        self.addCleanup(delete_files, 'only_variables', directory=folder_nnp, exceptions=['only_variables.in'])

    def test_nnp_mkdir(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        sweep = Sweep({},fullpath = fullpath)
        sweep.config.set('nextnano++','outputdirectory',r'tests//outputs')
        created_directory = os.path.join(sweep.mk_dir(overwrite=True))
        self.assertTrue(os.path.samefile(created_directory, os.path.join('tests', 'outputs', 'only_variables_sweep')))
        self.assertTrue(os.path.isdir(os.path.join('tests', 'outputs', 'only_variables_sweep')))

        self.addCleanup(os.rmdir, os.path.join('tests', 'outputs', 'only_variables_sweep'))
        self.addCleanup(delete_files, 'only_variables', directory=folder_nnp, exceptions=['only_variables.in'])

    def test_nnp_mkdir_specify(self):
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        sweep = Sweep({}, fullpath=fullpath)
        created_directory = os.path.join(sweep.mk_dir(overwrite=True, output_directory=r'tests//outputs'))
        self.assertTrue(os.path.samefile(created_directory, os.path.join('tests', 'outputs', 'only_variables_sweep')))
        self.assertTrue(os.path.isdir(os.path.join('tests', 'outputs', 'only_variables_sweep')))

        self.addCleanup(os.rmdir, os.path.join('tests', 'outputs', 'only_variables_sweep'))
        self.addCleanup(delete_files, 'only_variables', directory=folder_nnp, exceptions=['only_variables.in'])

    def test_nnp_save(self):
        self.addCleanup(delete_files, 'only_variables', directory=folder_nnp, exceptions=['only_variables.in'])
        fullpath = os.path.join(folder_nnp, 'only_variables.in')
        sweep = Sweep({'float': [1, 2], 'str': ['test1', 'test2']}, fullpath)
        sweep.save_sweep()

        files_with_names = [file for file in os.listdir(folder_nnp) if 'only_variables' in file]
        self.assertEqual(len(files_with_names), 5)
        self.assertTrue(os.path.isfile(os.path.join(folder_nnp, 'only_variables__float_2_str_test1_.in')))




    #nn3 section
    def test_nn3_init(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        self.assertRaises(ValueError, Sweep, {'Name': 'some_name'}, fullpath)

        sweep = Sweep({},fullpath)
        self.assertEqual(sweep.fullpath, fullpath)
        self.assertEqual(sweep.var_sweep, {})
        self.assertFalse(sweep.input_files)
        self.assertFalse(sweep.sweep_output_directory)


        self.assertRaises(ValueError, Sweep, {'float':1})
        self.assertRaises(TypeError, Sweep, {'float': 1}, fullpath)

        sweep = Sweep({'float': [1,2,5]}, fullpath)
        self.assertEqual(sweep.fullpath, fullpath)
        self.assertEqual(sweep.var_sweep['float'],[1,2,5])

        self.addCleanup(delete_files, 'only_variables', directory=folder_nn3, exceptions=['only_variables.in'])

    def test_nn3_mkdir(self):
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        sweep = Sweep({},fullpath = fullpath)
        sweep.config.set('nextnano3','outputdirectory',r'tests//outputs')
        #self.assertEqual(os.path.normpath(os.path.join(sweep.mk_dir(overwrite=True))), os.path.normpath(r'tests//outputs\only_variables_sweep'))
        created_directory = os.path.join(sweep.mk_dir(overwrite=True))
        self.assertTrue(os.path.samefile(created_directory,os.path.join('tests','outputs','only_variables_sweep')))
        self.assertTrue(os.path.isdir(os.path.join('tests','outputs', 'only_variables_sweep')))

        self.addCleanup(os.rmdir,os.path.join('tests','outputs', 'only_variables_sweep'))
        self.addCleanup(delete_files, 'only_variables', directory=folder_nn3, exceptions=['only_variables.in'])

    def test_nn3_save(self):
        self.addCleanup(delete_files, 'only_variables', directory=folder_nn3, exceptions=['only_variables.in'])
        fullpath = os.path.join(folder_nn3, 'only_variables.in')
        sweep = Sweep({'float': [1, 2], 'str': ['test1', 'test2']}, fullpath)
        sweep.save_sweep()

        files_with_names = [file for file in os.listdir(folder_nn3) if 'only_variables' in file]
        self.assertEqual(len(files_with_names), 5)
        self.assertTrue(os.path.isfile(os.path.join(folder_nn3, 'only_variables__float_2_str_test1_.in')))

    #TODO test parallel sweeps with and without convergenceCheck
    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        if os.path.isdir(r'tests//outputs\only_variables_sweep'):
            os.rmdir(r'tests//outputs\only_variables_sweep')
        delete_files('only_variables', directory=folder_nnp, exceptions=['only_variables.in'])
        delete_files('only_variables', directory=folder_nn3, exceptions=['only_variables.in'])




if __name__ == '__main__':
    unittest.main()

    file = os.path.join(folder_nnp, 'only_variables.in')
    fi = InputFile(file)
    text = fi.text



from nextnanopy.utils.formatting import str_to_path, _path, pattern_in_file, is_variable, parse_variable, \
    generate_command, pattern_in_text
from collections import OrderedDict

fmt = {
    'var_char': 'NOT DEFINED',
    'com_char': '<!--',
    'input_pattern': '<nextnano.MSB',
}

config_validator = {
    'exe': str_to_path,
    'license': str_to_path,
    'database': str_to_path,
    'outputdirectory': str_to_path,
    'debug': int,
}

config_default = {
    'exe': '',
    'license': '',
    'database': '',
    'outputdirectory': '',
    'debug': 0,
}


def command_msb(
        inputfile,
        exe,
        license,
        database,
        outputdirectory,
        debug=0,
        **kwargs,
):
    kwargs = OrderedDict(
        exe=[_path(exe), ''],
        inputfile=['-inputfile', _path(inputfile)],
        license=['-license', _path(license)],
        database=['-database', _path(database)],
        outputdirectory=['-outputdirectory', _path(outputdirectory)],
    )
    if debug == 1:
        kwargs['debug'] = ['-debug', debug]
    return generate_command(kwargs.values())


def is_msb_variable(text):
    return is_variable(text, var_char=fmt['var_char'])


def parse_msb_variable(text):
    return parse_variable(text, var_char=fmt['var_char'], com_char=fmt['com_char'])


def is_msb_input_file(fullpath):
    return pattern_in_file(fullpath, fmt['input_pattern'])


def is_msb_input_text(text):
    return pattern_in_text(text, fmt['input_pattern'])



from nextnanopy.outputs import DataFolder, DataFile
from nextnanopy.utils.misc import combinations
from collections.abc import Iterable
import os
import numpy as np
from scipy.optimize import minimize as sp_minimize
from scipy.optimize import fsolve

elementary_charge = 1.60217662e-19

def minimization_function(variables_values, inputfile, variables_names, target_filepath, target_variable, number =0, overwrite = True):
    for varvalue, varname in zip(variables_values, variables_names):
        inputfile.set_variable(name = varname, value = varvalue)
    inputfile.save(overwrite = True)
    inputfile.execute(show_log = False)
    output_folder = inputfile.folder_output
    datafolder = DataFolder(output_folder)
    datafile = DataFile(datafolder.go_to(os.path.join(target_filepath)))
    target = datafile.variables[target_variable].value[number]
    return target

def optimization_function(variables_values, inputfile, variables_names, target_filepath, target_variable, target_value, number =0, overwrite = True):
    for varvalue, varname in zip(variables_values, variables_names):
        inputfile.set_variable(name = varname, value = varvalue)
    inputfile.save(overwrite = True)
    inputfile.execute(show_log = False)
    output_folder = inputfile.folder_output
    datafolder = DataFolder(output_folder)
    datafile = DataFile(datafolder.go_to(os.path.join(target_filepath)), product = 'nextnano++')
    target = datafile.variables[target_variable].value[number]
    target = abs(target_value - target)
    return target

def minimize(input_file, variables_names, x0, target_filepath, target_variable,  number = 0, overwrite = True):
    if isinstance(variables_names, Iterable):
        if len(variables_names) != len(x0):
            raise ValueError('variables and x0 should be of the same length')
    result = sp_minimize(minimization_function, x0, args = (input_file, variables_names, target_filepath, target_variable, number, overwrite), tol = 1e-5)
    return result



def optimize(input_file, variables_names, x0, target_filepath, target_variable, target_value, number = 0, overwrite = True):
    if isinstance(variables_names,Iterable):
        if len(variables_names) != len(x0):
            raise ValueError('variables and x0 should be of the same length')
    result = fsolve(optimization_function, x0, args = (input_file, variables_names, target_filepath, target_variable, target_value, number, overwrite), xtol = 1e-5,factor=1)
    input_file.save(overwrite = True)
    return result





def get_target_value(input_file,target_filepath,target_variable,target_number = 0, product = 'nextnano++'):
    dfolder_output = DataFolder(input_file.folder_output)
    dfile_output = dfolder_output.go_to(target_filepath)
    dfile = DataFile(dfile_output, product = product)
    if target_number == 'all':
        val = dfile.variables[target_variable].value
    else:
        val = dfile.variables[target_variable].value[target_number]
    return val


def simple_optimize(input_file, sweep_dict, target_filepath, target_variable, target_number, goal = 'min', optimal_target_value = None, post_func = lambda x: x):
    """
    The function that run sweeps of input file over the given variables space
    and finds optimum value of target value (target value is part of the result of simulation)
    Parameters
    ----------
    input_file: str, input_file to execute
    sweep_dict: dict with keys: variabels names
                          values: list of variables values
    target_filepath: filepath to target value in output directory
    target_variable: target variable name
    target_number: the index number of the target value. If target_number = "all", the target is then list of values
    goal: str
            options: 'min' to minimize the target
                      'max' to maximize the target
                       'optimal' find the closest to optimal_target_value
    optimal_target_value
    post_func: if specifide post_func(target) will be optimize instead of target

    Returns
    -------

    """
    best_value_initial_dict = {'min': np.inf, 'max': -np.inf, 'optimal':0}
    var_names = sweep_dict.keys()
    var_values = sweep_dict.values()
    all_combinations = combinations(*var_values)
    optimal_combination = None
    best_value = best_value_initial_dict[goal]
    best_diff = np.inf
    for combination in all_combinations:
        for var_name,var_val in zip(var_names,combination):
            input_file.set_variable(var_name,var_val)
        input_file.save(overwrite = True)
        input_file.execute(show_log = False)
        target_val = post_func(get_target_value(input_file,target_filepath,target_variable,target_number))
        if goal == 'min':
            if target_val<best_value:
                best_value = target_val
                optimal_combination = combination
        elif goal == 'max':
            if target_val>best_value:
                best_value = target_val
                optimal_combination = combination
        elif goal == 'optimal':
            diff = np.abs(target_val - optimal_target_value)
            if diff < best_diff:
                best_value = target_val
                optimal_combination = combination
                best_diff = diff
        else:
            raise NotImplementedError
    return optimal_combination, best_value

def calculate_CV(output_directory_path, bias1 = None, bias2 = None, total = False, net_charge_sign = -1):
    """
    Calculates CV charateristic based on integrated_density_electron.dat and integrated_density_hole.dat files in ouput_directory
    Parameters
    ----------
    output_directory_path: str
    bias1:  str, name of the first reference bias, optional
    bias2: str, name of the second reference bias, optional
    total: calculate in all regions together (sum of all regions) (for now only False is valid, otherwise return NotImplementedError)

    Returns
    -------
    capacitance: numpy array with capacitance
    voltage: numpy array with voltages
    """

    dfolder = DataFolder(output_directory_path)

    dfile_hole  = DataFile(dfolder.file('integrated_density_hole.dat'), FirstVarIsCoordFlag=False, product = 'nextnano++')
    dfile_electron = DataFile(dfolder.file('integrated_density_electron.dat'), FirstVarIsCoordFlag=False, product = 'nextnano++')

    bias_hole_set = set([var.name for var in dfile_hole.variables if var.name.endswith('_bias')])
    bias_electron_set = set([var.name for var in dfile_electron.variables if var.name.endswith('_bias')])
    common_biases_list = list(bias_hole_set.intersection(bias_electron_set))

    """
    Here parameter voltage is found.
    if both bias1 and bias2 are present in holes and electrones file, voltage = bias2-bias1
    
    bias 2 is ignored if bias1 is None
    
    if bias1 is present in common_biases_list, voltage = bias1
    
    if bias1 = bias2 = None, voltage = first common bias
    """
    if not common_biases_list:
        raise ValueError('There is no common biases in integrated_density_hole.dat file and integrated_density_electron.dat file')

    if bias1:
        if bias1 not in common_biases_list:
            raise ValueError('Specified reference bias1 is not commmon bias for holes and electrons')

        if bias2:
            if bias2 not in common_biases_list:
                raise ValueError('Specified reference bias2 is not commmon bias for holes and electrons')
            else:
                voltage = dfile_hole.variables[bias2].value - dfile_hole.variables[bias1].value
        else:
            voltage = dfile_hole.variables[bias1].value

    else:
        bias = common_biases_list[0]
        voltage = dfile_hole.variables[bias].value

    regions_hole_set = set([var.name for var in dfile_hole.variables if not var.name.endswith('_bias')])
    regions_electron_set = set([var.name for var in dfile_electron.variables if not var.name.endswith('_bias')])



    common_regions_list = list(regions_hole_set.intersection(regions_electron_set))

    if not common_regions_list:
        return TypeError('integrated_density does not have common regions for holes and electrons')

    #voltage = voltage[:-1]

    num_bias_points = len(voltage)

    regions = []

    for region in common_regions_list:
        integrated_density_hole = dfile_hole[region].value
        integrated_density_electron = dfile_electron[region].value
        total_charge = net_charge_sign * elementary_charge * 1e6 * (integrated_density_hole-integrated_density_electron)
        c_list = []
        regions.append(c_list)

        for i in range(len(voltage)-1):
            voltage_val = voltage[i]
            voltage_val_next = voltage[i+1]
            charge_val = total_charge[i]
            charge_val_next = total_charge[i+1]

            if abs(voltage_val_next - voltage_val) < 1e-15:
                c_list.append(0)
            else:
                c_val = (charge_val_next-charge_val)/(voltage_val_next-voltage_val)
                c_list.append(c_val)

    voltage = voltage[:-1]

    return voltage, regions







